@@ -20,8 +20,8 @@
 #include <contrib/cleanup.h>
 #include <libknot/descriptor.h>
 
-#include "lib/cache.h"
-#include "lib/cdb.h"
+#include "lib/cache/api.h"
+#include "lib/cache/cdb_api.h"
 #include "lib/utils.h"
 #include "daemon/bindings.h"
 #include "daemon/worker.h"
@@ -85,7 +85,7 @@ static int mod_load(lua_State *L)
 		lua_error(L);
 	}
 	/* Parse precedence declaration */
-	auto_free char *declaration = strdup(lua_tostring(L, 1));
+	char *declaration = strdup(lua_tostring(L, 1));
 	if (!declaration) {
 		return kr_error(ENOMEM);
 	}
@@ -95,6 +95,7 @@ static int mod_load(lua_State *L)
 	/* Load engine module */
 	struct engine *engine = engine_luaget(L);
 	int ret = engine_register(engine, name, precedence, ref);
+	free(declaration);
 	if (ret != 0) {
 		if (ret == kr_error(EIDRM)) {
 			format_error(L, "referenced module not found");
@@ -390,7 +391,7 @@ static int net_tls(lua_State *L)
 
 	int r = tls_certificate_set(net, lua_tostring(L, 1), lua_tostring(L, 2));
 	if (r != 0) {
-		lua_pushstring(L, strerror(ENOMEM));
+		lua_pushstring(L, kr_strerror(r));
 		lua_error(L);
 	}
 
@@ -398,6 +399,187 @@ static int net_tls(lua_State *L)
 	return 1;
 }
 
+static int print_tls_param(const char *key, void *val, void *data)
+{
+	if (!val) {
+		return 0;
+	}
+
+	struct tls_client_paramlist_entry *entry = (struct tls_client_paramlist_entry *)val;
+
+	lua_State *L = (lua_State *)data;
+
+	lua_createtable(L, 0, 3);
+
+	lua_createtable(L, entry->pins.len, 0);
+	for (size_t i = 0; i < entry->pins.len; ++i) {
+		lua_pushnumber(L, i + 1);
+		lua_pushstring(L, entry->pins.at[i]);
+		lua_settable(L, -3);
+	}
+	lua_setfield(L, -2, "pins");
+
+	lua_createtable(L, entry->ca_files.len, 0);
+	for (size_t i = 0; i < entry->ca_files.len; ++i) {
+		lua_pushnumber(L, i + 1);
+		lua_pushstring(L, entry->ca_files.at[i]);
+		lua_settable(L, -3);
+	}
+	lua_setfield(L, -2, "ca_files");
+
+	lua_createtable(L, entry->hostnames.len, 0);
+	for (size_t i = 0; i < entry->hostnames.len; ++i) {
+		lua_pushnumber(L, i + 1);
+		lua_pushstring(L, entry->hostnames.at[i]);
+		lua_settable(L, -3);
+	}
+	lua_setfield(L, -2, "hostnames");
+
+	lua_setfield(L, -2, key);
+
+	return 0;
+}
+
+static int print_tls_client_params(lua_State *L)
+{
+	struct engine *engine = engine_luaget(L);
+	if (!engine) {
+		return 0;
+	}
+	struct network *net = &engine->net;
+	if (!net) {
+		return 0;
+	}
+	if (net->tls_client_params.root == 0 ) {
+		return 0;
+	}
+	lua_newtable(L);
+	map_walk(&net->tls_client_params, print_tls_param, (void *)L);
+	return 1;
+}
+
+
+static int net_tls_client(lua_State *L)
+{
+	struct engine *engine = engine_luaget(L);
+	if (!engine) {
+		return 0;
+	}
+	struct network *net = &engine->net;
+	if (!net) {
+		return 0;
+	}
+
+	/* Only return current credentials. */
+	if (lua_gettop(L) == 0) {
+		return print_tls_client_params(L);
+	}
+
+	const char *full_addr = NULL;
+	bool pin_exists = false;
+	bool ca_file_exists = false;
+	if ((lua_gettop(L) == 1) && lua_isstring(L, 1)) {
+		full_addr = lua_tostring(L, 1);
+	} else if ((lua_gettop(L) == 2) && lua_isstring(L, 1) && lua_istable(L, 2)) {
+		full_addr = lua_tostring(L, 1);
+		pin_exists = true;
+	} else if ((lua_gettop(L) == 3) && lua_isstring(L, 1) && lua_istable(L, 2)) {
+		full_addr = lua_tostring(L, 1);
+		ca_file_exists = true;
+	} else if ((lua_gettop(L) == 4) && lua_isstring(L, 1) &&
+		    lua_istable(L, 2) && lua_istable(L, 3)) {
+		full_addr = lua_tostring(L, 1);
+		pin_exists = true;
+		ca_file_exists = true;
+	} else {
+		format_error(L, "net.tls_client takes one parameter (\"address\"), two parameters (\"address\",\"pin\"), three parameters (\"address\", \"ca_file\", \"hostname\") or four ones: (\"address\", \"pin\", \"ca_file\", \"hostname\")");
+		lua_error(L);
+	}
+
+	char addr[INET6_ADDRSTRLEN];
+	uint16_t port = 0;
+	if (kr_straddr_split(full_addr, addr, sizeof(addr), &port) != kr_ok()) {
+		format_error(L, "invalid IP address");
+		lua_error(L);
+	}
+
+	if (port == 0) {
+		port = 853;
+	}
+
+	if (!pin_exists && !ca_file_exists) {
+		int r = tls_client_params_set(&net->tls_client_params,
+					      addr, port, NULL, NULL, NULL);
+		if (r != 0) {
+			lua_pushstring(L, kr_strerror(r));
+			lua_error(L);
+		}
+
+		lua_pushboolean(L, true);
+		return 1;
+	}
+
+	if (pin_exists) {
+		/* iterate over table with pins
+		 * http://www.lua.org/manual/5.1/manual.html#lua_next */
+		lua_pushnil(L); /* first key */
+		while (lua_next(L, 2)) {  /* pin table is in stack at index 2 */
+			/* pin now at index -1, key at index -2*/
+			const char *pin = lua_tostring(L, -1);
+			int r = tls_client_params_set(&net->tls_client_params,
+						      addr, port, NULL, NULL, pin);
+			if (r != 0) {
+				lua_pushstring(L, kr_strerror(r));
+				lua_error(L);
+			}
+			lua_pop(L, 1);
+		}
+	}
+
+	int ca_table_index = 2;
+	int hostname_table_index = 3;
+	if (ca_file_exists) {
+		if (pin_exists) {
+			ca_table_index = 3;
+			hostname_table_index = 4;
+		}
+	} else {
+		lua_pushboolean(L, true);
+		return 1;
+	}
+
+	/* iterate over ca filenames */
+	lua_pushnil(L);
+	while (lua_next(L, ca_table_index)) {
+		const char *ca_file = lua_tostring(L, -1);
+		int r = tls_client_params_set(&net->tls_client_params,
+					      addr, port, ca_file, NULL, NULL);
+		if (r != 0) {
+			lua_pushstring(L, kr_strerror(r));
+			lua_error(L);
+		}
+		/* removes 'value'; keeps 'key' for next iteration */
+		lua_pop(L, 1);
+	}
+
+	/* iterate over hostnames */
+	lua_pushnil(L);
+	while (lua_next(L, hostname_table_index)) {
+		const char *hostname = lua_tostring(L, -1);
+		int r = tls_client_params_set(&net->tls_client_params,
+					      addr, port, NULL, hostname, NULL);
+		if (r != 0) {
+			lua_pushstring(L, kr_strerror(r));
+			lua_error(L);
+		}
+		/* removes 'value'; keeps 'key' for next iteration */
+		lua_pop(L, 1);
+	}
+
+	lua_pushboolean(L, true);
+	return 1;
+}
+
 static int net_tls_padding(lua_State *L)
 {
 	struct engine *engine = engine_luaget(L);
@@ -508,6 +690,8 @@ int lib_net(lua_State *L)
 		{ "bufsize",      net_bufsize },
 		{ "tcp_pipeline", net_pipeline },
 		{ "tls",          net_tls },
+		{ "tls_server",   net_tls },
+		{ "tls_client",   net_tls_client },
 		{ "tls_padding",  net_tls_padding },
 		{ "outgoing_v4",  net_outgoing_v4 },
 		{ "outgoing_v6",  net_outgoing_v6 },
@@ -547,6 +731,33 @@ static int cache_count(lua_State *L)
 	return 0;
 }
 
+/** Return time of last checkpoint, or re-set it if passed `true`. */
+static int cache_checkpoint(lua_State *L)
+{
+	struct engine *engine = engine_luaget(L);
+	struct kr_cache *cache = &engine->resolver.cache;
+
+	if (lua_gettop(L) == 0) { /* Return the current value. */
+		lua_newtable(L);
+		lua_pushnumber(L, cache->checkpoint_monotime);
+		lua_setfield(L, -2, "monotime");
+		lua_newtable(L);
+		lua_pushnumber(L, cache->checkpoint_walltime.tv_sec);
+		lua_setfield(L, -2, "sec");
+		lua_pushnumber(L, cache->checkpoint_walltime.tv_usec);
+		lua_setfield(L, -2, "usec");
+		lua_setfield(L, -2, "walltime");
+		return 1;
+	}
+
+	if (lua_gettop(L) != 1 || !lua_isboolean(L, 1) || !lua_toboolean(L, 1)) {
+		format_error(L, "cache.checkpoint() takes no parameters or a true value");
+		lua_error(L);
+	}
+	kr_cache_make_checkpoint(cache);
+	return 1;
+}
+
 /** Return cache statistics. */
 static int cache_stats(lua_State *L)
 {
@@ -668,8 +879,12 @@ static int cache_open(lua_State *L)
 	};
 	int ret = kr_cache_open(&engine->resolver.cache, api, &opts, engine->pool);
 	if (ret != 0) {
-		format_error(L, "can't open cache");
-		lua_error(L);
+		char cwd[PATH_MAX];
+		if(getcwd(cwd, sizeof(cwd)) == NULL) {
+			const char errprefix[] = "<invalid working directory>";
+			strncpy(cwd, errprefix, sizeof(cwd));
+		}
+		return luaL_error(L, "can't open cache path '%s'; working directory '%s'", opts.path, cwd);
 	}
 
 	/* Store current configuration */
@@ -704,12 +919,13 @@ static int cache_close(lua_State *L)
 	return 1;
 }
 
+#if 0
 /** @internal Prefix walk. */
 static int cache_prefixed(struct kr_cache *cache, const char *args, knot_db_val_t *results, int maxresults)
 {
 	/* Decode parameters */
 	uint8_t namespace = 'R';
-	char *extra = (char *)strchr(args, ' ');
+	char *extra = strchr(args, ' ');
 	if (extra != NULL) {
 		extra[0] = '\0';
 		namespace = extra[1];
@@ -722,7 +938,9 @@ static int cache_prefixed(struct kr_cache *cache, const char *args, knot_db_val_
 	}
 
 	/* Start prefix search */
-	return kr_cache_match(cache, namespace, buf, results, maxresults);
+	int ret = kr_cache_match(cache, namespace, buf, results, maxresults);
+	kr_cache_sync(cache);
+	return ret;
 }
 
 /** @internal Delete iterated key. */
@@ -748,12 +966,14 @@ static int cache_remove_prefix(struct kr_cache *cache, const char *args)
 		result_set[i].data = dst;
 	}
 	cache->api->remove(cache->db, result_set, ret);
+	kr_cache_sync(cache);
 	/* Free keys */
 	for (int i = 0; i < ret; ++i) {
 		free(result_set[i].data);
 	}
 	return ret;
 }
+#endif
 
 /** Prune expired/invalid records. */
 static int cache_prune(lua_State *L)
@@ -803,7 +1023,7 @@ static int cache_clear(lua_State *L)
 
 	/* Clear a sub-tree in cache. */
 	if (args && strlen(args) > 0) {
-		int ret = cache_remove_prefix(cache, args);
+		int ret = kr_error(ENOSYS); // FIXME cache_remove_prefix(cache, args);
 		if (ret < 0) {
 			format_error(L, kr_strerror(ret));
 			lua_error(L);
@@ -881,11 +1101,10 @@ static int cache_get(lua_State *L)
 		lua_error(L);
 	}
 
-	/* Clear a sub-tree in cache. */
-	const char *args = lua_tostring(L, 1);
 	/* Retrieve set of keys */
+	//const char *args = lua_tostring(L, 1);
 	static knot_db_val_t result_set[100];
-	int ret = cache_prefixed(cache, args, result_set, 100);
+	int ret = kr_error(ENOSYS); // FIXME cache_prefixed(cache, args, result_set, 100);
 	if (ret < 0) {
 		format_error(L, kr_strerror(ret));
 		lua_error(L);
@@ -904,6 +1123,7 @@ int lib_cache(lua_State *L)
 		{ "backends", cache_backends },
 		{ "count",  cache_count },
 		{ "stats",  cache_stats },
+		{ "checkpoint", cache_checkpoint },
 		{ "open",   cache_open },
 		{ "close",  cache_close },
 		{ "prune",  cache_prune },
@@ -1160,70 +1380,82 @@ int lib_event(lua_State *L)
 	return 1;
 }
 
-/* @internal Call the Lua callback stored in baton. */
-static void resolve_callback(struct worker_ctx *worker, struct kr_request *req, void *baton)
-{
-	assert(worker);
-	assert(req);
-	assert(baton);
-	lua_State *L = worker->engine->L;
-	intptr_t cb_ref = (intptr_t) baton;
-	lua_rawgeti(L, LUA_REGISTRYINDEX, cb_ref);
-	luaL_unref(L, LUA_REGISTRYINDEX, cb_ref);
-	lua_pushlightuserdata(L, req->answer);
-	lua_pushlightuserdata(L, req);
-	(void) execute_callback(L, 2);
-}
-
 static int wrk_resolve(lua_State *L)
 {
 	struct worker_ctx *worker = wrk_luaget(L);
 	if (!worker) {
 		return 0;
 	}
-	/* Create query packet */
-	knot_pkt_t *pkt = knot_pkt_new(NULL, KNOT_EDNS_MAX_UDP_PAYLOAD, NULL);
-	if (!pkt) {
-		lua_pushstring(L, strerror(ENOMEM));
-		lua_error(L);
-	}
+
 	uint8_t dname[KNOT_DNAME_MAXLEN];
 	if (!knot_dname_from_str(dname, lua_tostring(L, 1), sizeof(dname))) {
 		lua_pushstring(L, "invalid qname");
 		lua_error(L);
 	};
+
 	/* Check class and type */
 	uint16_t rrtype = lua_tointeger(L, 2);
 	if (!lua_isnumber(L, 2)) {
 		lua_pushstring(L, "invalid RR type");
 		lua_error(L);
 	}
+
 	uint16_t rrclass = lua_tointeger(L, 3);
 	if (!lua_isnumber(L, 3)) { /* Default class is IN */
 		rrclass = KNOT_CLASS_IN;
 	}
+
+	/* Add query options */
+	const struct kr_qflags *options = lua_topointer(L, 4);
+	if (!options) { /* but we rely on the lua wrapper when dereferencing non-NULL */
+		lua_pushstring(L, "invalid options");
+		lua_error(L);
+	}
+
+	/* Create query packet */
+	knot_pkt_t *pkt = knot_pkt_new(NULL, KNOT_EDNS_MAX_UDP_PAYLOAD, NULL);
+	if (!pkt) {
+		lua_pushstring(L, kr_strerror(ENOMEM));
+		lua_error(L);
+	}
 	knot_pkt_put_question(pkt, dname, rrclass, rrtype);
 	knot_wire_set_rd(pkt->wire);
+
 	/* Add OPT RR */
 	pkt->opt_rr = knot_rrset_copy(worker->engine->resolver.opt_rr, NULL);
 	if (!pkt->opt_rr) {
+		knot_pkt_free(&pkt);
 		return kr_error(ENOMEM);
-	}	
-	/* Add completion callback */
-	int ret = 0;
-	unsigned options = lua_tointeger(L, 4);
-	if (lua_isfunction(L, 5)) {
-		/* Store callback in registry */
-		lua_pushvalue(L, 5);
-		int cb = luaL_ref(L, LUA_REGISTRYINDEX);
-		ret = worker_resolve(worker, pkt, options, resolve_callback, (void *) (intptr_t)cb);
-	} else {
-		ret = worker_resolve(worker, pkt, options, NULL, NULL);
 	}
-	
+	if (options->DNSSEC_WANT) {
+		knot_edns_set_do(pkt->opt_rr);
+	}
+
+	if (options->DNSSEC_CD) {
+		knot_wire_set_cd(pkt->wire);
+	}
+
+	/* Create task and start with a first question */
+	struct qr_task *task = worker_resolve_start(worker, pkt, *options);
+	if (!task) {
+		knot_rrset_free(&pkt->opt_rr, NULL);
+		knot_pkt_free(&pkt);
+		lua_pushstring(L, "couldn't create a resolution request");
+		lua_error(L);
+	}
+
+	/* Add initialisation callback */
+	if (lua_isfunction(L, 5)) {
+		lua_pushvalue(L, 5);
+		lua_pushlightuserdata(L, worker_task_request(task));
+		(void) execute_callback(L, 1);
+	}
+
+	/* Start execution */
+	int ret = worker_resolve_exec(task, pkt);
+	lua_pushboolean(L, ret == 0);
 	knot_rrset_free(&pkt->opt_rr, NULL);
 	knot_pkt_free(&pkt);
-	lua_pushboolean(L, ret == 0);
 	return 1;
 }
 
@@ -1282,7 +1514,7 @@ static int wrk_stats(lua_State *L)
 int lib_worker(lua_State *L)
 {
 	static const luaL_Reg lib[] = {
-		{ "resolve",  wrk_resolve },
+		{ "resolve_unwrapped",  wrk_resolve },
 		{ "stats",    wrk_stats },
 		{ NULL, NULL }
 	};
