@@ -28,9 +28,9 @@
 #include "daemon/bindings.h"
 #include "daemon/ffimodule.h"
 #include "lib/nsrep.h"
-#include "lib/cache.h"
+#include "lib/cache/api.h"
 #include "lib/defines.h"
-#include "lib/cdb_lmdb.h"
+#include "lib/cache/cdb_lmdb.h"
 #include "lib/dnssec/ta.h"
 
 /** @internal Compatibility wrapper for Lua < 5.2 */
@@ -232,9 +232,10 @@ int engine_set_moduledir(struct engine *engine, const char *moduledir) {
 	/* Save original package.path to package._path */
 	snprintf(l_paths, MAXPATHLEN - 1,
 		 "if package._path == nil then package._path = package.path end\n"
-		 "package.path = '%s/?.lua;%s/?/init.lua;'..package._path\n",
-		 new_moduledir,
-		 new_moduledir);
+		 "package.path = '%1$s/?.lua;%1$s/?/init.lua;'..package._path\n"
+		 "if package._cpath == nil then package._cpath = package.cpath end\n"
+		 "package.cpath = '%1$s/?%2$s;'..package._cpath\n",
+		 new_moduledir, LIBEXT);
 
 	int ret = l_dobytecode(engine->L, l_paths, strlen(l_paths), "");
 	if (ret != 0) {
@@ -266,37 +267,6 @@ static int l_moduledir(lua_State *L)
 	return 1;
 }
 
-/** Get/set context option. */
-static int l_option(lua_State *L)
-{
-	struct engine *engine = engine_luaget(L);
-	/* Look up option name */
-	unsigned opt_code = 0;
-	if (lua_isstring(L, 1)) {
-		const char *opt = lua_tostring(L, 1);
-		for (const knot_lookup_t *it = kr_query_flag_names(); it->name; ++it) {
-			if (strcmp(it->name, opt) == 0) {
-				opt_code = it->id;
-				break;
-			}
-		}
-		if (!opt_code) {
-			lua_pushstring(L, "invalid option name");
-			lua_error(L);
-		}
-	}
-	/* Get or set */
-	if (lua_isboolean(L, 2) || lua_isnumber(L, 2)) {
-		if (lua_toboolean(L, 2)) {
-			engine->resolver.options |= opt_code;
-		} else {
-			engine->resolver.options &= ~opt_code; 
-		}
-	}
-	lua_pushboolean(L, engine->resolver.options & opt_code);
-	return 1;
-}
-
 /** @internal for l_trustanchor: */
 static void ta_add(zs_scanner_t *zs)
 {
@@ -349,6 +319,66 @@ static int l_trustanchor(lua_State *L)
 	lua_pushboolean(L, true);
 	return 1;
 }
+
+/** Load root hints from zonefile. */
+static int l_hint_root_file(lua_State *L)
+{
+	struct engine *engine = engine_luaget(L);
+	struct kr_context *ctx = &engine->resolver;
+	const char *file = lua_tostring(L, 1);
+
+	const char *err = engine_hint_root_file(ctx, file);
+	if (err) {
+		if (!file) {
+			file = ROOTHINTS;
+		}
+		lua_push_printf(L, "error when opening '%s': %s", file, err);
+		lua_error(L);
+	} else {
+		lua_pushboolean(L, true);
+		return 1;
+	}
+}
+
+/** @internal for engine_hint_root_file */
+static void roothints_add(zs_scanner_t *zs)
+{
+	struct kr_zonecut *hints = zs->process.data;
+	if (!hints) {
+		return;
+	}
+	if (zs->r_type == KNOT_RRTYPE_A || zs->r_type == KNOT_RRTYPE_AAAA) {
+		knot_rdata_t rdata[RDATA_ARR_MAX];
+		knot_rdata_init(rdata, zs->r_data_length, zs->r_data, zs->r_ttl);
+		kr_zonecut_add(hints, zs->r_owner, rdata);
+	}
+}
+const char* engine_hint_root_file(struct kr_context *ctx, const char *file)
+{
+	if (!file) {
+		file = ROOTHINTS;
+	}
+	if (strlen(file) == 0 || !ctx) {
+		return "invalid parameters";
+	}
+	struct kr_zonecut *root_hints = &ctx->root_hints;
+
+	zs_scanner_t zs;
+	if (zs_init(&zs, ".", 1, 0) != 0) {
+		return "not enough memory";
+	}
+	if (zs_set_input_file(&zs, file) != 0) {
+		zs_deinit(&zs);
+		return "failed to open root hints file";
+	}
+
+	kr_zonecut_set(root_hints, (const uint8_t *)"");
+	zs_set_processing(&zs, roothints_add, NULL, root_hints);
+	zs_parse_all(&zs);
+	zs_deinit(&zs);
+	return NULL;
+}
+
 /** Unpack JSON object to table */
 static void l_unpack_json(lua_State *L, JsonNode *table)
 {
@@ -574,8 +604,7 @@ static int init_resolver(struct engine *engine)
 	/* Load basic modules */
 	engine_register(engine, "iterate", NULL, NULL);
 	engine_register(engine, "validate", NULL, NULL);
-	engine_register(engine, "rrcache", NULL, NULL);
-	engine_register(engine, "pktcache", NULL, NULL);
+	engine_register(engine, "cache", NULL, NULL);
 
 	return array_push(engine->backends, kr_cdb_lmdb());
 }
@@ -601,12 +630,12 @@ static int init_state(struct engine *engine)
 	lua_setglobal(engine->L, "moduledir");
 	lua_pushcfunction(engine->L, l_verbose);
 	lua_setglobal(engine->L, "verbose");
-	lua_pushcfunction(engine->L, l_option);
-	lua_setglobal(engine->L, "option");
 	lua_pushcfunction(engine->L, l_setuser);
 	lua_setglobal(engine->L, "user");
 	lua_pushcfunction(engine->L, l_trustanchor);
 	lua_setglobal(engine->L, "trustanchor");
+	lua_pushcfunction(engine->L, l_hint_root_file);
+	lua_setglobal(engine->L, "_hint_root_file");
 	lua_pushliteral(engine->L, libknot_SONAME);
 	lua_setglobal(engine->L, "libknot_SONAME");
 	lua_pushliteral(engine->L, libzscanner_SONAME);
@@ -633,6 +662,34 @@ static void update_state(uv_timer_t *handle)
 	lru_apply(engine->resolver.cache_rtt, update_stat_item, NULL);
 }
 
+/**
+ * Start luacov measurement and store results to file specified by
+ * KRESD_COVERAGE_STATS environment variable.
+ * Do nothing if the variable is not set.
+ */
+static void init_measurement(struct engine *engine)
+{
+	const char * const statspath = getenv("KRESD_COVERAGE_STATS");
+	if (!statspath)
+		return;
+
+	char * snippet = NULL;
+	int ret = asprintf(&snippet,
+		"_luacov_runner = require('luacov.runner')\n"
+		"_luacov_runner.init({\n"
+		"	statsfile = '%s',\n"
+		"	exclude = {'test', 'tapered', 'lua/5.1'},\n"
+		"})\n"
+		"jit.off()\n", statspath
+	);
+	assert(ret > 0);
+
+	ret = luaL_loadstring(engine->L, snippet);
+	assert(ret == 0);
+	lua_call(engine->L, 0, 0);
+	free(snippet);
+}
+
 int engine_init(struct engine *engine, knot_mm_t *pool)
 {
 	if (engine == NULL) {
@@ -646,7 +703,9 @@ int engine_init(struct engine *engine, knot_mm_t *pool)
 	int ret = init_state(engine);
 	if (ret != 0) {
 		engine_deinit(engine);
+		return ret;
 	}
+	init_measurement(engine);
 	/* Initialize resolver */
 	ret = init_resolver(engine);
 	if (ret != 0) {
@@ -664,8 +723,8 @@ static void engine_unload(struct engine *engine, struct kr_module *module)
 	/* Unregister module */
 	auto_free char *name = strdup(module->name);
 	kr_module_unload(module);
-	/* Clear in Lua world */
-	if (name) {
+	/* Clear in Lua world, but not for embedded modules ('cache' in particular). */
+	if (name && !kr_module_embedded(name)) {
 		lua_pushnil(engine->L);
 		lua_setglobal(engine->L, name);
 	}
@@ -752,9 +811,8 @@ int engine_ipc(struct engine *engine, const char *expr)
 	}
 }
 
-static int engine_loadconf(struct engine *engine, const char *config_path)
+int engine_load_sandbox(struct engine *engine)
 {
-	int ret = 0;
 	/* Init environment */
 	static const char sandbox_bytecode[] = {
 		#include "daemon/lua/sandbox.inc"
@@ -764,22 +822,13 @@ static int engine_loadconf(struct engine *engine, const char *config_path)
 		lua_pop(engine->L, 1);
 		return kr_error(ENOEXEC);
 	}
-	/* Load config file */
-	if (config_path) {
-		if (strcmp(config_path, "-") == 0) {
-			return ret; /* No config and no defaults. */
-		}
-		ret = l_dosandboxfile(engine->L, config_path);
-	}
-	if (ret == 0) {
-		/* Load defaults */
-		static const char config_bytecode[] = {
-			#include "daemon/lua/config.inc"
-		};
-		ret = l_dobytecode(engine->L, config_bytecode, sizeof(config_bytecode), "config");
-	}
+	return kr_ok();
+}
 
-	/* Evaluate */
+int engine_loadconf(struct engine *engine, const char *config_path)
+{
+	assert(config_path != NULL);
+	int ret = l_dosandboxfile(engine->L, config_path);
 	if (ret != 0) {
 		fprintf(stderr, "%s\n", lua_tostring(engine->L, -1));
 		lua_pop(engine->L, 1);
@@ -787,14 +836,22 @@ static int engine_loadconf(struct engine *engine, const char *config_path)
 	return ret;
 }
 
-int engine_start(struct engine *engine, const char *config_path)
+int engine_load_defaults(struct engine *engine)
 {
-	/* Load configuration. */
-	int ret = engine_loadconf(engine, config_path);
+	/* Load defaults */
+	static const char config_bytecode[] = {
+		#include "daemon/lua/config.inc"
+	};
+	int ret = l_dobytecode(engine->L, config_bytecode, sizeof(config_bytecode), "config");
 	if (ret != 0) {
-		return ret;
+		fprintf(stderr, "%s\n", lua_tostring(engine->L, -1));
+		lua_pop(engine->L, 1);
 	}
+	return ret;
+}
 
+int engine_start(struct engine *engine)
+{
 	/* Clean up stack and restart GC */
 	lua_settop(engine->L, 0);
 	lua_gc(engine->L, LUA_GCCOLLECT, 0);
