@@ -2,6 +2,8 @@
 
 import sys, os, re, subprocess
 
+import sphinx_rtd_theme
+
 # -- General configuration -----------------------------------------------------
 
 if os.environ.get('READTHEDOCS', None) == 'True':
@@ -42,15 +44,12 @@ primary_domain = 'c'
 # so a file named "default.css" will overwrite the builtin "default.css".
 html_static_path = ['_static']
 
-# If true, SmartyPants will be used to convert quotes and dashes to
-# typographically correct entities.
-html_use_smartypants = True
-
 # Output file base name for HTML help builder.
 htmlhelp_basename = 'apidoc'
 
 # Theme
 html_theme = 'sphinx_rtd_theme'
+html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
 
 # -- Options for LaTeX output --------------------------------------------------
 
