@@ -43,6 +43,7 @@ void kr_crypto_reinit(void);
 struct dseckey;
 
 #define KR_DNSSEC_VFLG_WEXPAND 0x01
+#define KR_DNSSEC_VFLG_OPTOUT  0x02
 
 /** DNSSEC validation context. */
 struct kr_rrset_validation_ctx {
@@ -70,23 +71,13 @@ typedef struct kr_rrset_validation_ctx kr_rrset_validation_ctx_t;
 int kr_rrset_validate(kr_rrset_validation_ctx_t *vctx,
 			const knot_rrset_t *covered);
 
-/**
- * Validate RRSet using a specific key.
- * @param vctx    Pointer to validation context.
- * @param covered RRSet covered by a signature. It must be in canonical format.
- * @param key_pos Position of the key to be validated with.
- * @param key     Key to be used to validate.
- *		  If NULL, then key from DNSKEY RRSet is used.
- * @return        0 or error code, same as vctx->result.
- */
-int kr_rrset_validate_with_key(kr_rrset_validation_ctx_t *vctx,
-				const knot_rrset_t *covered,
-				size_t key_pos, const struct dseckey *key);
 /**
  * Check whether the DNSKEY rrset matches the supplied trust anchor RRSet.
- * @param vctx Pointer to validation context.
- * @param ta   Trust anchor RRSet against which to validate the DNSKEY RRSet.
- * @return     0 or error code, same as vctx->result.
+ * @param vctx  Pointer to validation context.
+ * @param ta    Trust anchor RRSet against which to validate the DNSKEY RRSet.
+ * @return      0 or error code, same as vctx->result.  In particular,
+ * 		DNSSEC_INVALID_DS_ALGORITHM if *each* DS records is unusable
+ * 		due to unimplemented DNSKEY or DS algorithm.
  */
 int kr_dnskeys_trusted(kr_rrset_validation_ctx_t *vctx, const knot_rrset_t *ta);
 
@@ -129,6 +120,7 @@ int kr_dnssec_key_match(const uint8_t *key_a_rdata, size_t key_a_rdlen,
  * @param kown  DNSKEY owner name.
  * @param rdata DNSKEY RDATA
  * @param rdlen DNSKEY RDATA length
+ * @return 0 or error code; in particular: DNSSEC_INVALID_KEY_ALGORITHM
  */
 int kr_dnssec_key_from_rdata(struct dseckey **key, const knot_dname_t *kown, const uint8_t *rdata, size_t rdlen);
 
@@ -137,3 +129,14 @@ int kr_dnssec_key_from_rdata(struct dseckey **key, const knot_dname_t *kown, con
  * @param key Pointer to freed key.
  */
 void kr_dnssec_key_free(struct dseckey **key);
+
+/**
+ * Checks whether NSEC/NSEC3 RR selected by iterator matches the supplied name and type.
+ * @param rrs     Records selected by iterator.
+ * @param qry_uid Query unique identifier where NSEC/NSEC3 belongs to.
+ * @param name    Name to be checked.
+ * @param type    Type to be checked.
+ * @return        0 or error code.
+ */
+int kr_dnssec_matches_name_and_type(const ranked_rr_array_t *rrs, uint32_t qry_uid,
+				    const knot_dname_t *name, uint16_t type);
