@@ -31,6 +31,7 @@
 #include <assert.h>
 #include <arpa/inet.h>
 
+#include <contrib/cleanup.h>
 #include <libknot/descriptor.h>
 #include <libknot/rrtype/rdname.h>
 #include <libknot/rrtype/rrsig.h>
@@ -54,7 +55,7 @@ static const knot_dname_t *minimized_qname(struct kr_query *query, uint16_t *qty
 {
 	/* Minimization disabled. */
 	const knot_dname_t *qname = query->sname;
-	if (qname[0] == '\0' || query->flags & (QUERY_NO_MINIMIZE|QUERY_STUB)) {
+	if (qname[0] == '\0' || query->flags.NO_MINIMIZE || query->flags.STUB) {
 		return qname;
 	}
 
@@ -111,7 +112,7 @@ static bool is_authoritative(const knot_pkt_t *answer, struct kr_query *query)
 	return false;
 }
 
-int kr_response_classify(knot_pkt_t *pkt)
+int kr_response_classify(const knot_pkt_t *pkt)
 {
 	const knot_pktsection_t *an = knot_pkt_section(pkt, KNOT_ANSWER);
 	switch (knot_wire_get_rcode(pkt->wire)) {
@@ -147,7 +148,7 @@ static bool is_valid_addr(const uint8_t *addr, size_t len)
 }
 
 /** @internal Update NS address from record \a rr.  Return _FAIL on error. */
-static int update_nsaddr(const knot_rrset_t *rr, struct kr_query *query)
+static int update_nsaddr(const knot_rrset_t *rr, struct kr_query *query, int *glue_cnt)
 {
 	if (rr->type == KNOT_RRTYPE_A || rr->type == KNOT_RRTYPE_AAAA) {
 		const knot_rdata_t *rdata = rr->rrs.data;
@@ -155,13 +156,13 @@ static int update_nsaddr(const knot_rrset_t *rr, struct kr_query *query)
 		const int addr_len = knot_rdata_rdlen(rdata);
 		char name_str[KNOT_DNAME_MAXLEN];
 		char addr_str[INET6_ADDRSTRLEN];
-		WITH_VERBOSE {
+		WITH_VERBOSE(query) {
 			const int af = (addr_len == sizeof(struct in_addr)) ?
 				       AF_INET : AF_INET6;
 			knot_dname_to_str(name_str, rr->owner, sizeof(name_str));
 			inet_ntop(af, addr, addr_str, sizeof(addr_str));
 		}
-		if (!(query->flags & QUERY_ALLOW_LOCAL) &&
+		if (!(query->flags.ALLOW_LOCAL) &&
 			!is_valid_addr(addr, addr_len)) {
 			QVERBOSE_MSG(query, "<= ignoring invalid glue for "
 				     "'%s': '%s'\n", name_str, addr_str);
@@ -171,15 +172,23 @@ static int update_nsaddr(const knot_rrset_t *rr, struct kr_query *query)
 		if (ret != 0) {
 			return KR_STATE_FAIL;
 		}
-		QVERBOSE_MSG(query, "<= using glue for "
+
+		++*glue_cnt; /* reduced verbosity */
+		/* QVERBOSE_MSG(query, "<= using glue for "
 			     "'%s': '%s'\n", name_str, addr_str);
+		*/
 	}
 	return KR_STATE_CONSUME;
 }
 
-/** @internal From \a pkt, fetch glue records for name \a ns, and update the cut etc. */
-static void fetch_glue(knot_pkt_t *pkt, const knot_dname_t *ns, struct kr_request *req)
+/** @internal From \a pkt, fetch glue records for name \a ns, and update the cut etc.
+ *
+ * \param glue_cnt the number of accepted addresses (to be incremented)
+ */
+static void fetch_glue(knot_pkt_t *pkt, const knot_dname_t *ns, bool in_bailiwick,
+			struct kr_request *req, const struct kr_query *qry, int *glue_cnt)
 {
+	ranked_rr_array_t *selected[] = kr_request_selected(req);
 	for (knot_section_t i = KNOT_ANSWER; i <= KNOT_ADDITIONAL; ++i) {
 		const knot_pktsection_t *sec = knot_pkt_section(pkt, i);
 		for (unsigned k = 0; k < sec->count; ++k) {
@@ -191,21 +200,27 @@ static void fetch_glue(knot_pkt_t *pkt, const knot_dname_t *ns, struct kr_reques
 			    (rr->type != KNOT_RRTYPE_AAAA)) {
 				continue;
 			}
+
+			uint8_t rank = (in_bailiwick && i == KNOT_ANSWER)
+				? (KR_RANK_INITIAL | KR_RANK_AUTH) : KR_RANK_OMIT;
+			(void) kr_ranked_rrarray_add(selected[i], rr, rank,
+							false, qry->uid, &req->pool);
+
 			if ((rr->type == KNOT_RRTYPE_A) &&
-			    (req->ctx->options & QUERY_NO_IPV4)) {
+			    (req->ctx->options.NO_IPV4)) {
 				continue;
 			}
 			if ((rr->type == KNOT_RRTYPE_AAAA) &&
-			    (req->ctx->options & QUERY_NO_IPV6)) {
+			    (req->ctx->options.NO_IPV6)) {
 				continue;
 			}
-			(void) update_nsaddr(rr, req->current_query);
+			(void) update_nsaddr(rr, req->current_query, glue_cnt);
 		}
 	}
 }
 
 /** Attempt to find glue for given nameserver name (best effort). */
-static int has_glue(knot_pkt_t *pkt, const knot_dname_t *ns)
+static bool has_glue(knot_pkt_t *pkt, const knot_dname_t *ns)
 {
 	for (knot_section_t i = KNOT_ANSWER; i <= KNOT_ADDITIONAL; ++i) {
 		const knot_pktsection_t *sec = knot_pkt_section(pkt, i);
@@ -213,18 +228,19 @@ static int has_glue(knot_pkt_t *pkt, const knot_dname_t *ns)
 			const knot_rrset_t *rr = knot_pkt_rr(sec, k);
 			if (knot_dname_is_equal(ns, rr->owner) &&
 			    (rr->type == KNOT_RRTYPE_A || rr->type == KNOT_RRTYPE_AAAA)) {
-				return 1;
+				return true;
 			}
 		}
 	}
-	return 0;
+	return false;
 }
 
 /** @internal Update the cut with another NS(+glue) record.
  * @param current_cut is cut name before this packet.
  * @return _DONE if cut->name changes, _FAIL on error, and _CONSUME otherwise. */
 static int update_cut(knot_pkt_t *pkt, const knot_rrset_t *rr,
-		      struct kr_request *req, const knot_dname_t *current_cut)
+		      struct kr_request *req, const knot_dname_t *current_cut,
+		      int *glue_cnt)
 {
 	struct kr_query *qry = req->current_query;
 	struct kr_zonecut *cut = &qry->zone_cut;
@@ -267,24 +283,33 @@ static int update_cut(knot_pkt_t *pkt, const knot_rrset_t *rr,
 	/* Fetch glue for each NS */
 	for (unsigned i = 0; i < rr->rrs.rr_count; ++i) {
 		const knot_dname_t *ns_name = knot_ns_name(&rr->rrs, i);
-		int glue_records = has_glue(pkt, ns_name);
 		/* Glue is mandatory for NS below zone */
-		if (!glue_records && knot_dname_in(rr->owner, ns_name)) {
-			VERBOSE_MSG("<= authority: missing mandatory glue, rejecting\n");
+		if (knot_dname_in(rr->owner, ns_name) && !has_glue(pkt, ns_name)) {
+			const char *msg =
+				"<= authority: missing mandatory glue, skipping NS";
+			WITH_VERBOSE(qry) {
+				auto_free char *ns_str = kr_dname_text(ns_name);
+				VERBOSE_MSG("%s %s\n", msg, ns_str);
+			}
 			continue;
 		}
-		kr_zonecut_add(cut, ns_name, NULL);
+		int ret = kr_zonecut_add(cut, ns_name, NULL);
+		assert(!ret); (void)ret;
+
 		/* Choose when to use glue records. */
-		if (qry->flags & QUERY_PERMISSIVE) {
-			fetch_glue(pkt, ns_name, req);
-		} else if (qry->flags & QUERY_STRICT) {
+		bool in_bailiwick = knot_dname_in(current_cut, ns_name);
+		bool do_fetch;
+		if (qry->flags.PERMISSIVE) {
+			do_fetch = true;
+		} else if (qry->flags.STRICT) {
 			/* Strict mode uses only mandatory glue. */
-			if (knot_dname_in(cut->name, ns_name))
-				fetch_glue(pkt, ns_name, req);
+			do_fetch = knot_dname_in(cut->name, ns_name);
 		} else {
 			/* Normal mode uses in-bailiwick glue. */
-			if (knot_dname_in(current_cut, ns_name))
-				fetch_glue(pkt, ns_name, req);
+			do_fetch = in_bailiwick;
+		}
+		if (do_fetch) {
+			fetch_glue(pkt, ns_name, in_bailiwick, req, qry, glue_cnt);
 		}
 	}
 
@@ -292,35 +317,26 @@ static int update_cut(knot_pkt_t *pkt, const knot_rrset_t *rr,
 }
 
 /** Compute rank appropriate for RRs present in the packet.
- * @param answer whether the RR is from answer or authority section */
+ * @param answer whether the RR is from answer or authority section
+ * @param is_nonauth: from referral or forwarding (etc.) */
 static uint8_t get_initial_rank(const knot_rrset_t *rr, const struct kr_query *qry,
-				const bool answer, const bool is_referral)
+				const bool answer, const bool is_nonauth)
 {
 	/* For RRSIGs, ensure the KR_RANK_AUTH flag corresponds to the signed RR. */
 	uint16_t type = kr_rrset_type_maysig(rr);
 
-	if (qry->flags & QUERY_CACHED) {
+	if (qry->flags.CACHED) {
 		return rr->additional ? *(uint8_t *)rr->additional : KR_RANK_OMIT;
 		/* ^^ Current use case for "cached" RRs without rank: hints module. */
 	}
 	if (answer || type == KNOT_RRTYPE_DS
+	    || type == KNOT_RRTYPE_SOA /* needed for aggressive negative caching */
 	    || type == KNOT_RRTYPE_NSEC || type == KNOT_RRTYPE_NSEC3) {
+		/* We almost always want these validated, and it should be possible. */
 		return KR_RANK_INITIAL | KR_RANK_AUTH;
 	}
-	if (type == KNOT_RRTYPE_NS) {
-		/* Some servers add extra NS RRset, which allows us to refresh
-		 * cache "for free", potentially speeding up zone cut lookups
-		 * in future.  Still, it might theoretically cause some problems:
-		 * https://mailarchive.ietf.org/arch/msg/dnsop/CYjPDlwtpxzdQV_qycB-WfnW6CI
-		 */
-		if (!is_referral && knot_dname_is_equal(qry->zone_cut.name, rr->owner)) {
-			return KR_RANK_INITIAL | KR_RANK_AUTH;
-		} else {
-			return KR_RANK_OMIT;
-		}
-	}
-
-	return KR_RANK_INITIAL;
+	/* Be aggressive: try to validate anything else (almost never extra latency). */
+	return KR_RANK_TRY;
 	/* TODO: this classifier of authoritativity may not be perfect yet. */
 }
 
@@ -343,7 +359,8 @@ static int pick_authority(knot_pkt_t *pkt, struct kr_request *req, bool to_wire)
 		if (!knot_dname_in(zonecut_name, rr->owner)) {
 			continue;
 		}
-		uint8_t rank = get_initial_rank(rr, qry, false, referral);
+		uint8_t rank = get_initial_rank(rr, qry, false,
+						qry->flags.FORWARD || referral);
 		int ret = kr_ranked_rrarray_add(&req->auth_selected, rr,
 						rank, to_wire, qry->uid, &req->pool);
 		if (ret != kr_ok()) {
@@ -357,10 +374,10 @@ static int pick_authority(knot_pkt_t *pkt, struct kr_request *req, bool to_wire)
 static int process_authority(knot_pkt_t *pkt, struct kr_request *req)
 {
 	struct kr_query *qry = req->current_query;
-	assert(!(qry->flags & QUERY_STUB));
+	assert(!(qry->flags.STUB));
 
 	int result = KR_STATE_CONSUME;
-	if (qry->flags & QUERY_FORWARD) {
+	if (qry->flags.FORWARD) {
 		return result;
 	}
 
@@ -392,12 +409,13 @@ static int process_authority(knot_pkt_t *pkt, struct kr_request *req)
 	/* Remember current bailiwick for NS processing. */
 	const knot_dname_t *current_zone_cut = qry->zone_cut.name;
 	bool ns_record_exists = false;
+	int glue_cnt = 0;
 	/* Update zone cut information. */
 	for (unsigned i = 0; i < ns->count; ++i) {
 		const knot_rrset_t *rr = knot_pkt_rr(ns, i);
 		if (rr->type == KNOT_RRTYPE_NS) {
 			ns_record_exists = true;
-			int state = update_cut(pkt, rr, req, current_zone_cut);
+			int state = update_cut(pkt, rr, req, current_zone_cut, &glue_cnt);
 			switch(state) {
 			case KR_STATE_DONE: result = state; break;
 			case KR_STATE_FAIL: return state; break;
@@ -409,8 +427,12 @@ static int process_authority(knot_pkt_t *pkt, struct kr_request *req)
 		}
 	}
 
+	if (glue_cnt) {
+		VERBOSE_MSG("<= loaded %d glue addresses\n", glue_cnt);
+	}
 
-	if ((qry->flags & QUERY_DNSSEC_WANT) && (result == KR_STATE_CONSUME)) {
+
+	if ((qry->flags.DNSSEC_WANT) && (result == KR_STATE_CONSUME)) {
 		if (knot_wire_get_aa(pkt->wire) == 0 &&
 		    knot_wire_get_ancount(pkt->wire) == 0 &&
 		    ns_record_exists) {
@@ -435,7 +457,7 @@ static void finalize_answer(knot_pkt_t *pkt, struct kr_query *qry, struct kr_req
 static int unroll_cname(knot_pkt_t *pkt, struct kr_request *req, bool referral, const knot_dname_t **cname_ret)
 {
 	struct kr_query *query = req->current_query;
-	assert(!(query->flags & QUERY_STUB));
+	assert(!(query->flags.STUB));
 	/* Process answer type */
 	const knot_pktsection_t *an = knot_pkt_section(pkt, KNOT_ANSWER);
 	const knot_dname_t *cname = NULL;
@@ -443,7 +465,7 @@ static int unroll_cname(knot_pkt_t *pkt, struct kr_request *req, bool referral,
 	unsigned cname_chain_len = 0;
 	bool is_final = (query->parent == NULL);
 	uint32_t iter_count = 0;
-	bool strict_mode = (query->flags & QUERY_STRICT);
+	bool strict_mode = (query->flags.STRICT);
 	do {
 		/* CNAME was found at previous iteration, but records may not follow the correct order.
 		 * Try to find records for pending_cname owner from section start. */
@@ -471,7 +493,7 @@ static int unroll_cname(knot_pkt_t *pkt, struct kr_request *req, bool referral,
 					continue;
 				}
 				if (rrsig_labels < cname_labels) {
-					query->flags |= QUERY_DNSSEC_WEXPAND;
+					query->flags.DNSSEC_WEXPAND = true;
 				}
 			}
 
@@ -482,12 +504,14 @@ static int unroll_cname(knot_pkt_t *pkt, struct kr_request *req, bool referral,
 				/* if not referral, mark record to be written to final answer */
 				to_wire = !referral;
 			} else {
-				state = update_nsaddr(rr, query->parent);
+				int cnt_ = 0;
+				state = update_nsaddr(rr, query->parent, &cnt_);
 				if (state == KR_STATE_FAIL) {
 					return state;
 				}
 			}
-			uint8_t rank = get_initial_rank(rr, query, true, referral);
+			uint8_t rank = get_initial_rank(rr, query, true,
+					query->flags.FORWARD || referral);
 			state = kr_ranked_rrarray_add(&req->answ_selected, rr,
 						      rank, to_wire, query->uid, &req->pool);
 			if (state != kr_ok()) {
@@ -549,7 +573,7 @@ static int process_referral_answer(knot_pkt_t *pkt, struct kr_request *req)
 		return KR_STATE_FAIL;
 	}
 	struct kr_query *query = req->current_query;
-	if (!(query->flags & QUERY_CACHED)) {
+	if (!(query->flags.CACHED)) {
 		/* If not cached (i.e. got from upstream)
 		 * make sure that this is not an authoritative answer
 		 * (even with AA=1) for other layers.
@@ -616,13 +640,13 @@ static int process_answer(knot_pkt_t *pkt, struct kr_request *req)
 	if (!knot_dname_is_equal(knot_pkt_qname(pkt), query->sname) &&
 	    (pkt_class & (PKT_NOERROR|PKT_NXDOMAIN|PKT_REFUSED|PKT_NODATA))) {
 		VERBOSE_MSG("<= found cut, retrying with non-minimized name\n");
-		query->flags |= QUERY_NO_MINIMIZE;
+		query->flags.NO_MINIMIZE = true;
 		return KR_STATE_CONSUME;
 	}
 
 	/* This answer didn't improve resolution chain, therefore must be authoritative (relaxed to negative). */
 	if (!is_authoritative(pkt, query)) {
-		if (!(query->flags & QUERY_FORWARD) &&
+		if (!(query->flags.FORWARD) &&
 		    pkt_class & (PKT_NXDOMAIN|PKT_NODATA)) {
 			VERBOSE_MSG("<= lame response: non-auth sent negative response\n");
 			return KR_STATE_FAIL;
@@ -638,17 +662,17 @@ static int process_answer(knot_pkt_t *pkt, struct kr_request *req)
 	/* Make sure that this is an authoritative answer (even with AA=0) for other layers */
 	knot_wire_set_aa(pkt->wire);
 	/* Either way it resolves current query. */
-	query->flags |= QUERY_RESOLVED;
+	query->flags.RESOLVED = true;
 	/* Follow canonical name as next SNAME. */
 	if (!knot_dname_is_equal(cname, query->sname)) {
 		/* Check if target record has been already copied */
-		query->flags |= QUERY_CNAME;
+		query->flags.CNAME = true;
 		if (is_final) {
 			state = process_final(pkt, req, cname);
 			if (state != kr_ok()) {
 				return state;
 			}
-		} else if ((query->flags & QUERY_FORWARD) &&
+		} else if ((query->flags.FORWARD) &&
 			   ((query->stype == KNOT_RRTYPE_DS) ||
 			    (query->stype == KNOT_RRTYPE_NS))) {
 			/* CNAME'ed answer for DS or NS subquery.
@@ -670,9 +694,9 @@ static int process_answer(knot_pkt_t *pkt, struct kr_request *req)
 		if (!next) {
 			return KR_STATE_FAIL;
 		}
-		next->flags |= QUERY_AWAIT_CUT;
-		if (query->flags & QUERY_FORWARD) {
-			next->forward_flags |= QUERY_CNAME;
+		next->flags.AWAIT_CUT = true;
+		if (query->flags.FORWARD) {
+			next->forward_flags.CNAME = true;
 			if (query->parent == NULL) {
 				state = kr_nsrep_copy_set(&next->ns, &query->ns);
 				if (state != kr_ok()) {
@@ -684,12 +708,12 @@ static int process_answer(knot_pkt_t *pkt, struct kr_request *req)
 		/* Want DNSSEC if and only if it's posible to secure
 		 * this name (i.e. iff it is covered by a TA) */
 		if (kr_ta_covers_qry(req->ctx, cname, query->stype)) {
-			next->flags |= QUERY_DNSSEC_WANT;
+			next->flags.DNSSEC_WANT = true;
 		} else {
-			next->flags &= ~QUERY_DNSSEC_WANT;
+			next->flags.DNSSEC_WANT = false;
 		}
-		if (!(query->flags & QUERY_FORWARD) ||
-		    (query->flags & QUERY_DNSSEC_WEXPAND)) {
+		if (!(query->flags.FORWARD) ||
+		    (query->flags.DNSSEC_WEXPAND)) {
 			state = pick_authority(pkt, req, false);
 			if (state != kr_ok()) {
 				return KR_STATE_FAIL;
@@ -736,17 +760,17 @@ static int process_answer(knot_pkt_t *pkt, struct kr_request *req)
 	return KR_STATE_DONE;
 }
 
-/** @internal like process_answer() but for the forwarding mode. */
+/** @internal like process_answer() but for the STUB mode. */
 static int process_stub(knot_pkt_t *pkt, struct kr_request *req)
 {
 	struct kr_query *query = req->current_query;
-	assert(query->flags & QUERY_STUB);
+	assert(query->flags.STUB);
 	/* Pick all answer RRs. */
 	const knot_pktsection_t *an = knot_pkt_section(pkt, KNOT_ANSWER);
 	for (unsigned i = 0; i < an->count; ++i) {
 		const knot_rrset_t *rr = knot_pkt_rr(an, i);
 		int err = kr_ranked_rrarray_add(&req->answ_selected, rr,
-			      KR_RANK_INITIAL | KR_RANK_AUTH, true, query->uid, &req->pool);
+			      KR_RANK_OMIT | KR_RANK_AUTH, true, query->uid, &req->pool);
 		/* KR_RANK_AUTH: we don't have the records directly from
 		 * an authoritative source, but we do trust the server and it's
 		 * supposed to only send us authoritative records. */
@@ -756,7 +780,7 @@ static int process_stub(knot_pkt_t *pkt, struct kr_request *req)
 	}
 
 	knot_wire_set_aa(pkt->wire);
-	query->flags |= QUERY_RESOLVED;
+	query->flags.RESOLVED = true;
 	/* Pick authority RRs. */
 	int pkt_class = kr_response_classify(pkt);
 	const bool to_wire = ((pkt_class & (PKT_NXDOMAIN|PKT_NODATA)) != 0);
@@ -827,11 +851,11 @@ int kr_make_query(struct kr_query *query, knot_pkt_t *pkt)
 	query->id = rnd ^ (rnd >> 16); /* cheap way to strengthen unpredictability */
 	knot_wire_set_id(pkt->wire, query->id);
 	pkt->parsed = pkt->size;
-	WITH_VERBOSE {
+	WITH_VERBOSE(query) {
 		char name_str[KNOT_DNAME_MAXLEN], type_str[16];
 		knot_dname_to_str(name_str, query->sname, sizeof(name_str));
 		knot_rrtype_to_string(query->stype, type_str, sizeof(type_str));
-		QVERBOSE_MSG(query, "'%s' type '%s' id was assigned, parent id %hu\n",
+		QVERBOSE_MSG(query, "'%s' type '%s' id was assigned, parent id %u\n",
 			    name_str, type_str, query->parent ? query->parent->id : 0);
 	}
 	return kr_ok();
@@ -863,13 +887,13 @@ static int resolve_badmsg(knot_pkt_t *pkt, struct kr_request *req, struct kr_que
 
 #ifndef STRICT_MODE
 	/* Work around broken auths/load balancers */
-	if (query->flags & QUERY_SAFEMODE) {
+	if (query->flags.SAFEMODE) {
 		return resolve_error(pkt, req);
-	} else if (query->flags & QUERY_NO_MINIMIZE) {
-		query->flags |= QUERY_SAFEMODE;
+	} else if (query->flags.NO_MINIMIZE) {
+		query->flags.SAFEMODE = true;
 		return KR_STATE_DONE;
 	} else {
-		query->flags |= QUERY_NO_MINIMIZE;
+		query->flags.NO_MINIMIZE = true;
 		return KR_STATE_DONE;
 	}
 #else
@@ -890,14 +914,14 @@ static int resolve(kr_layer_t *ctx, knot_pkt_t *pkt)
 		return ctx->state;
 	}
 
-	WITH_VERBOSE {
-	if (query->flags & QUERY_TRACE) {
-		VERBOSE_MSG("<= answer received:\n");
-		kr_pkt_print(pkt);
-	}
+	WITH_VERBOSE(query) {
+		if (query->flags.TRACE) {
+			auto_free char *pkt_text = kr_pkt_text(pkt);
+			VERBOSE_MSG("<= answer received: \n%s\n", pkt_text);
+		}
 	}
 
-	if (query->flags & (QUERY_RESOLVED|QUERY_BADCOOKIE_AGAIN)) {
+	if (query->flags.RESOLVED || query->flags.BADCOOKIE_AGAIN) {
 		return ctx->state;
 	}
 
@@ -917,17 +941,17 @@ static int resolve(kr_layer_t *ctx, knot_pkt_t *pkt)
 		VERBOSE_MSG("<= ignoring mismatching response\n");
 		/* Force TCP, to work around authoritatives messing up question
 		 * without yielding to spoofed responses. */
-		query->flags |= QUERY_TCP;
+		query->flags.TCP = true;
 		return resolve_badmsg(pkt, req, query);
 	} else if (knot_wire_get_tc(pkt->wire)) {
 		VERBOSE_MSG("<= truncated response, failover to TCP\n");
 		if (query) {
 			/* Fail if already on TCP. */
-			if (query->flags & QUERY_TCP) {
+			if (query->flags.TCP) {
 				VERBOSE_MSG("<= TC=1 with TCP, bailing out\n");
 				return resolve_error(pkt, req);
 			}
-			query->flags |= QUERY_TCP;
+			query->flags.TCP = true;
 		}
 		return KR_STATE_CONSUME;
 	}
@@ -943,7 +967,7 @@ static int resolve(kr_layer_t *ctx, knot_pkt_t *pkt)
 		break; /* OK */
 	case KNOT_RCODE_REFUSED:
 	case KNOT_RCODE_SERVFAIL: {
-		if (query->flags & (QUERY_STUB | QUERY_FORWARD)) {
+		if (query->flags.STUB) {
 			 /* Pass through in stub mode */
 			break;
 		}
@@ -953,7 +977,9 @@ static int resolve(kr_layer_t *ctx, knot_pkt_t *pkt)
 			query->fails = 0; /* Reset per-query counter. */
 			return resolve_error(pkt, req);
 		} else {
-			query->flags |= QUERY_NO_MINIMIZE; /* Drop minimisation as a safe-guard. */
+			if (!query->flags.FORWARD) {
+				query->flags.NO_MINIMIZE = true; /* Drop minimisation as a safe-guard. */
+			}
 			return KR_STATE_CONSUME;
 		}
 	}
@@ -967,7 +993,7 @@ static int resolve(kr_layer_t *ctx, knot_pkt_t *pkt)
 	}
 
 	/* Forwarding/stub mode is special. */
-	if (query->flags & QUERY_STUB) {
+	if (query->flags.STUB) {
 		return process_stub(pkt, req);
 	}
 
