@@ -78,31 +78,18 @@ struct kr_nsrep
 	unsigned reputation;             /**< NS reputation */
 	const knot_dname_t *name;        /**< NS name */
 	struct kr_context *ctx;          /**< Resolution context */
-	union {
-		struct sockaddr ip;
-		struct sockaddr_in ip4;
-		struct sockaddr_in6 ip6;
-	} addr[KR_NSREP_MAXADDR];        /**< NS address(es) */
+	union inaddr addr[KR_NSREP_MAXADDR];        /**< NS address(es) */
 };
 
-/** @internal Address bytes for given family. */
-#define kr_nsrep_inaddr(addr) \
-	((addr).ip.sa_family == AF_INET ? (void *)&((addr).ip4.sin_addr) : (void *)&((addr).ip6.sin6_addr))
-/** @internal Address length for given family. */
-#define kr_nsrep_inaddr_len(addr) \
-	((addr).ip.sa_family == AF_INET ? sizeof(struct in_addr) : sizeof(struct in6_addr))
-
 /**
  * Set given NS address.
  * @param  qry      updated query
  * @param  index    index of the updated target
- * @param  addr     address bytes (struct in_addr or struct in6_addr)
- * @param  addr_len address bytes length (type will be derived from this)
- * @param  port     address port (if <= 0, 53 will be used)
+ * @param  sock     socket address to use (sockaddr_in or sockaddr_in6 or NULL)
  * @return          0 or an error code
  */
 KR_EXPORT
-int kr_nsrep_set(struct kr_query *qry, size_t index, uint8_t *addr, size_t addr_len, int port);
+int kr_nsrep_set(struct kr_query *qry, size_t index, const struct sockaddr *sock);
 
 /**
  * Elect best nameserver/address pair from the nsset.
@@ -148,3 +135,23 @@ int kr_nsrep_update_rtt(struct kr_nsrep *ns, const struct sockaddr *addr,
  */
 KR_EXPORT
 int kr_nsrep_update_rep(struct kr_nsrep *ns, unsigned reputation, kr_nsrep_lru_t *cache);
+/**
+ * Copy NSSET reputation information and resets score.
+ *
+ * @param  dst          updated NS representation
+ * @param  src          source NS representation
+ * @return              0 on success, error code on failure
+ */
+int kr_nsrep_copy_set(struct kr_nsrep *dst, const struct kr_nsrep *src);
+
+/**
+ * Sort addresses in the query nsrep list
+ * @param  ns           updated kr_nsrep
+ * @param  cache        RTT cache
+ * @return              0 or an error code
+ * @note   ns reputation is zeroed, as KR_NS_NOIP{4,6} flags are useless
+ * 		in STUB/FORWARD mode.
+ */
+KR_EXPORT
+int kr_nsrep_sort(struct kr_nsrep *ns, kr_nsrep_lru_t *cache);
+
