@@ -15,6 +15,7 @@
  */
 
 #include <ctype.h>
+#include <inttypes.h>
 #include <stdio.h>
 #include <fcntl.h>
 #include <assert.h>
@@ -43,6 +44,7 @@ bool kr_rank_check(uint8_t rank)
 	switch (rank & ~KR_RANK_AUTH) {
 	case KR_RANK_INITIAL:
 	case KR_RANK_OMIT:
+	case KR_RANK_TRY:
 	case KR_RANK_INDET:
 	case KR_RANK_BOGUS:
 	case KR_RANK_MISMATCH:
@@ -165,7 +167,9 @@ static int invalidate_ns(struct kr_rplan *rplan, struct kr_query *qry)
  */
 static void check_empty_nonterms(struct kr_query *qry, knot_pkt_t *pkt, struct kr_cache *cache, uint32_t timestamp)
 {
-	if (qry->flags & QUERY_NO_MINIMIZE) {
+	// FIXME cleanup, etc.
+#if 0
+	if (qry->flags.NO_MINIMIZE) {
 		return;
 	}
 
@@ -186,13 +190,15 @@ static void check_empty_nonterms(struct kr_query *qry, knot_pkt_t *pkt, struct k
 		int ret = kr_cache_peek(cache, KR_CACHE_PKT, target, KNOT_RRTYPE_NS, &entry, &timestamp);
 		if (ret == 0) { /* Either NXDOMAIN or NODATA, start here. */
 			/* @todo We could stop resolution here for NXDOMAIN, but we can't because of broken CDNs */
-			qry->flags |= QUERY_NO_MINIMIZE;
+			qry->flags.NO_MINIMIZE = true;
 			kr_make_query(qry, pkt);
-			return;
+			break;
 		}
 		assert(target[0]);
 		target = knot_wire_next_label(target, NULL);
 	}
+	kr_cache_sync(cache);
+#endif
 }
 
 static int ns_fetch_cut(struct kr_query *qry, const knot_dname_t *requested_name,
@@ -200,33 +206,35 @@ static int ns_fetch_cut(struct kr_query *qry, const knot_dname_t *requested_name
 {
 	/* It can occur that here parent query already have
 	 * provably insecured zonecut which not in the cache yet. */
-	const uint32_t insec_flags = QUERY_DNSSEC_INSECURE | QUERY_DNSSEC_NODS;
-	const uint32_t cut_flags = QUERY_AWAIT_IPV4 | QUERY_AWAIT_IPV6;
-	const bool is_insecured = ((qry->parent != NULL) &&
-				   (qry->parent->flags & cut_flags) == 0 &&
-				   (qry->parent->flags & insec_flags) != 0);
+	struct kr_qflags pflags;
+	if (qry->parent) {
+		pflags = qry->parent->flags;
+	}
+	const bool is_insecured = qry->parent != NULL
+		&& !(pflags.AWAIT_IPV4 || pflags.AWAIT_IPV6)
+		&& (pflags.DNSSEC_INSECURE || pflags.DNSSEC_NODS);
 
 	/* Want DNSSEC if it's possible to secure this name
 	 * (e.g. is covered by any TA) */
 	if (is_insecured) {
 		/* If parent is unsecured we don't want DNSSEC
 		 * even if cut name is covered by TA. */
-		qry->flags &= ~QUERY_DNSSEC_WANT;
-		qry->flags |= QUERY_DNSSEC_INSECURE;
+		qry->flags.DNSSEC_WANT = false;
+		qry->flags.DNSSEC_INSECURE = true;
 	} else if (kr_ta_covers_qry(req->ctx, qry->zone_cut.name, KNOT_RRTYPE_NS)) {
-		qry->flags |= QUERY_DNSSEC_WANT;
+		qry->flags.DNSSEC_WANT = true;
 	} else {
-		qry->flags &= ~QUERY_DNSSEC_WANT;
+		qry->flags.DNSSEC_WANT = false;
 	}
 
-	struct kr_zonecut cut_found = {0};
+	struct kr_zonecut cut_found;
 	kr_zonecut_init(&cut_found, requested_name, req->rplan.pool);
 	/* Cut that has been found can differs from cut that has been requested.
 	 * So if not already insecured,
 	 * try to fetch ta & keys even if initial cut name not covered by TA */
 	bool secured = !is_insecured;
 	int ret = kr_zonecut_find_cached(req->ctx, &cut_found, requested_name,
-					 qry->timestamp.tv_sec, &secured);
+					 qry, &secured);
 	if (ret == kr_error(ENOENT)) {
 		/* No cached cut found, start from SBELT
 		 * and issue priming query. */
@@ -235,7 +243,7 @@ static int ns_fetch_cut(struct kr_query *qry, const knot_dname_t *requested_name
 			return KR_STATE_FAIL;
 		}
 		VERBOSE_MSG(qry, "=> using root hints\n");
-		qry->flags &= ~QUERY_AWAIT_CUT;
+		qry->flags.AWAIT_CUT = false;
 		kr_zonecut_deinit(&cut_found);
 		return KR_STATE_DONE;
 	} else if (ret != kr_ok()) {
@@ -245,25 +253,24 @@ static int ns_fetch_cut(struct kr_query *qry, const knot_dname_t *requested_name
 
 	/* Find out security status.
 	 * Go insecure if the zone cut is provably insecure */
-	if ((qry->flags & QUERY_DNSSEC_WANT) && !secured) {
+	if ((qry->flags.DNSSEC_WANT) && !secured) {
 		VERBOSE_MSG(qry, "=> NS is provably without DS, going insecure\n");
-		qry->flags &= ~QUERY_DNSSEC_WANT;
-		qry->flags |= QUERY_DNSSEC_INSECURE;
+		qry->flags.DNSSEC_WANT = false;
+		qry->flags.DNSSEC_INSECURE = true;
 	}
 	/* Zonecut name can change, check it again
 	 * to prevent unnecessary DS & DNSKEY queries */
-	if (!(qry->flags & QUERY_DNSSEC_INSECURE) &&
+	if (!(qry->flags.DNSSEC_INSECURE) &&
 	    kr_ta_covers_qry(req->ctx, cut_found.name, KNOT_RRTYPE_NS)) {
-		qry->flags |= QUERY_DNSSEC_WANT;
+		qry->flags.DNSSEC_WANT = true;
 	} else {
-		qry->flags &= ~QUERY_DNSSEC_WANT;
+		qry->flags.DNSSEC_WANT = false;
 	}
 	/* Check if any DNSKEY found for cached cut */
-	if ((qry->flags & QUERY_DNSSEC_WANT) &&
-	    (cut_found.key == NULL)) {
+	if ((qry->flags.DNSSEC_WANT) && (cut_found.key == NULL)) {
 		/* No DNSKEY was found for cached cut.
 		 * If no glue were fetched for this cut,
-		 * we have got circular dependancy - must fetch A\AAAA
+		 * we have got circular dependency - must fetch A\AAAA
 		 * from authoritative, but we have no key to verify it.
 		 * TODO - try to refetch cut only if no glue were fetched */
 		kr_zonecut_deinit(&cut_found);
@@ -277,7 +284,7 @@ static int ns_fetch_cut(struct kr_query *qry, const knot_dname_t *requested_name
 			return KR_STATE_FAIL;
 		}
 		VERBOSE_MSG(qry, "=> using root hints\n");
-		qry->flags &= ~QUERY_AWAIT_CUT;
+		qry->flags.AWAIT_CUT = false;
 		return KR_STATE_DONE;
 	}
 	/* Copy fetched name */
@@ -305,14 +312,14 @@ static int ns_resolve_addr(struct kr_query *qry, struct kr_request *param)
 	 * Prefer IPv6 and continue with IPv4 if not available.
 	 */
 	uint16_t next_type = 0;
-	if (!(qry->flags & QUERY_AWAIT_IPV6) &&
-	    !(ctx->options & QUERY_NO_IPV6)) {
+	if (!(qry->flags.AWAIT_IPV6) &&
+	    !(ctx->options.NO_IPV6)) {
 		next_type = KNOT_RRTYPE_AAAA;
-		qry->flags |= QUERY_AWAIT_IPV6;
-	} else if (!(qry->flags & QUERY_AWAIT_IPV4) &&
-		   !(ctx->options & QUERY_NO_IPV4)) {
+		qry->flags.AWAIT_IPV6 = true;
+	} else if (!(qry->flags.AWAIT_IPV4) &&
+		   !(ctx->options.NO_IPV4)) {
 		next_type = KNOT_RRTYPE_A;
-		qry->flags |= QUERY_AWAIT_IPV4;
+		qry->flags.AWAIT_IPV4 = true;
 		/* Hmm, no useable IPv6 then. */
 		qry->ns.reputation |= KR_NS_NOIP6;
 		kr_nsrep_update_rep(&qry->ns, qry->ns.reputation, ctx->cache_rep);
@@ -323,10 +330,10 @@ static int ns_resolve_addr(struct kr_query *qry, struct kr_request *param)
 		if (!next_type && qry->zone_cut.name[0] == '\0') {
 			VERBOSE_MSG(qry, "=> fallback to root hints\n");
 			kr_zonecut_set_sbelt(ctx, &qry->zone_cut);
-			qry->flags |= QUERY_NO_THROTTLE; /* Pick even bad SBELT servers */
+			qry->flags.NO_THROTTLE = true; /* Pick even bad SBELT servers */
 			return kr_error(EAGAIN);
 		}
-		/* No IPv4 nor IPv6, flag server as unuseable. */
+		/* No IPv4 nor IPv6, flag server as unusable. */
 		VERBOSE_MSG(qry, "=> unresolvable NS address, bailing out\n");
 		qry->ns.reputation |= KR_NS_NOIP4 | KR_NS_NOIP6;
 		kr_nsrep_update_rep(&qry->ns, qry->ns.reputation, ctx->cache_rep);
@@ -336,10 +343,10 @@ static int ns_resolve_addr(struct kr_query *qry, struct kr_request *param)
 	struct kr_query *next = qry;
 	if (knot_dname_is_equal(qry->ns.name, qry->sname) &&
 	    qry->stype == next_type) {
-		if (!(qry->flags & QUERY_NO_MINIMIZE)) {
-			qry->flags |= QUERY_NO_MINIMIZE;
-			qry->flags &= ~QUERY_AWAIT_IPV6;
-			qry->flags &= ~QUERY_AWAIT_IPV4;
+		if (!(qry->flags.NO_MINIMIZE)) {
+			qry->flags.NO_MINIMIZE = true;
+			qry->flags.AWAIT_IPV6 = false;
+			qry->flags.AWAIT_IPV4 = false;
 			VERBOSE_MSG(qry, "=> circular dependepcy, retrying with non-minimized name\n");
 		} else {
 			qry->ns.reputation |= KR_NS_NOIP4 | KR_NS_NOIP6;
@@ -354,7 +361,7 @@ static int ns_resolve_addr(struct kr_query *qry, struct kr_request *param)
 		if (!next) {
 			return kr_error(ENOMEM);
 		}
-		next->flags |= QUERY_NONAUTH;
+		next->flags.NONAUTH = true;
 	}
 	/* At the root level with no NS addresses, add SBELT subrequest. */
 	int ret = 0;
@@ -363,10 +370,10 @@ static int ns_resolve_addr(struct kr_query *qry, struct kr_request *param)
 		if (ret == 0) { /* Copy TA and key since it's the same cut to avoid lookup. */
 			kr_zonecut_copy_trust(&next->zone_cut, &qry->zone_cut);
 			kr_zonecut_set_sbelt(ctx, &qry->zone_cut); /* Add SBELT to parent in case query fails. */
-			qry->flags |= QUERY_NO_THROTTLE; /* Pick even bad SBELT servers */
+			qry->flags.NO_THROTTLE = true; /* Pick even bad SBELT servers */
 		}
 	} else {
-		next->flags |= QUERY_AWAIT_CUT;
+		next->flags.AWAIT_CUT = true;
 	}
 	return ret;
 }
@@ -492,7 +499,7 @@ static int write_extra_ranked_records(const ranked_rr_array_t *arr, knot_pkt_t *
 				continue;
 			}
 		}
-		int err = knot_pkt_put(answer, 0, rr, 0);
+		err = knot_pkt_put(answer, 0, rr, 0);
 		if (err != KNOT_EOK) {
 			if (err == KNOT_ESPACE) {
 				err = kr_ok();
@@ -578,7 +585,7 @@ static int answer_finalize(struct kr_request *request, int state)
 	/* Always set SERVFAIL for bogus answers. */
 	if (state == KR_STATE_FAIL && rplan->pending.len > 0) {
 		struct kr_query *last = array_tail(rplan->pending);
-		if ((last->flags & QUERY_DNSSEC_WANT) && (last->flags & QUERY_DNSSEC_BOGUS)) {
+		if ((last->flags.DNSSEC_WANT) && (last->flags.DNSSEC_BOGUS)) {
 			return answer_fail(request);
 		}
 	}
@@ -590,10 +597,12 @@ static int answer_finalize(struct kr_request *request, int state)
 	 * Be conservative.  Primary approach: check ranks of all RRs in wire.
 	 * Only "negative answers" need special handling. */
 	bool secure = (last != NULL); /* suspicious otherwise */
-	if (last && (last->flags & QUERY_STUB)) {
+	VERBOSE_MSG(NULL, "AD: secure (start)\n");
+	if (last && (last->flags.STUB)) {
 		secure = false; /* don't trust forwarding for now */
 	}
-	if (last && (last->flags & QUERY_DNSSEC_OPTOUT)) {
+	if (last && (last->flags.DNSSEC_OPTOUT)) {
+		VERBOSE_MSG(NULL, "AD: opt-out\n");
 		secure = false; /* the last answer is insecure due to opt-out */
 	}
 
@@ -611,6 +620,7 @@ static int answer_finalize(struct kr_request *request, int state)
 		}
 	}
 
+	VERBOSE_MSG(NULL, "AD: secure (between ANS and AUTH)\n");
 	/* Write authority records. */
 	if (answer->current < KNOT_AUTHORITY) {
 		knot_pkt_begin(answer, KNOT_AUTHORITY);
@@ -637,13 +647,14 @@ static int answer_finalize(struct kr_request *request, int state)
 
 	/* AD: "negative answers" need more handling. */
 	if (last && secure) {
+		VERBOSE_MSG(NULL, "AD: secure (1)\n");
 		if (kr_response_classify(answer) != PKT_NOERROR
 		    /* Additionally check for CNAME chains that "end in NODATA",
 		     * as those would also be PKT_NOERROR. */
 		    || (answ_all_cnames && knot_pkt_qtype(answer) != KNOT_RRTYPE_CNAME))
 		{
-			secure = secure && (last->flags & QUERY_DNSSEC_WANT)
-				&& !(last->flags & (QUERY_DNSSEC_BOGUS | QUERY_DNSSEC_INSECURE));
+			secure = secure && last->flags.DNSSEC_WANT
+				&& !last->flags.DNSSEC_BOGUS && !last->flags.DNSSEC_INSECURE;
 		}
 	}
 	/* Clear AD if not secure.  ATM answer has AD=1 if requested secured answer. */
@@ -655,7 +666,7 @@ static int answer_finalize(struct kr_request *request, int state)
 	if (last) {
 		struct kr_query *cname_parent = last->cname_parent;
 		while (cname_parent != NULL) {
-			if (cname_parent->flags & QUERY_DNSSEC_OPTOUT) {
+			if (cname_parent->flags.DNSSEC_OPTOUT) {
 				knot_wire_clear_ad(answer->wire);
 				break;
 			}
@@ -670,7 +681,7 @@ static int query_finalize(struct kr_request *request, struct kr_query *qry, knot
 {
 	int ret = 0;
 	knot_pkt_begin(pkt, KNOT_ADDITIONAL);
-	if (!(qry->flags & QUERY_SAFEMODE)) {
+	if (!(qry->flags.SAFEMODE)) {
 		/* Remove any EDNS records from any previous iteration. */
 		ret = edns_erase_and_reserve(pkt);
 		if (ret == 0) {
@@ -678,7 +689,7 @@ static int query_finalize(struct kr_request *request, struct kr_query *qry, knot
 		}
 		if (ret == 0) {
 			/* Stub resolution (ask for +rd and +do) */
-			if (qry->flags & QUERY_STUB) {
+			if (qry->flags.STUB) {
 				knot_wire_set_rd(pkt->wire);
 				if (knot_pkt_has_dnssec(request->answer)) {
 					knot_edns_set_do(pkt->opt_rr);
@@ -687,11 +698,11 @@ static int query_finalize(struct kr_request *request, struct kr_query *qry, knot
 					knot_wire_set_cd(pkt->wire);
 				}
 			/* Full resolution (ask for +cd and +do) */
-			} else if (qry->flags & QUERY_FORWARD) {
+			} else if (qry->flags.FORWARD) {
 				knot_wire_set_rd(pkt->wire);
 				knot_edns_set_do(pkt->opt_rr);
 				knot_wire_set_cd(pkt->wire);
-			} else if (qry->flags & QUERY_DNSSEC_WANT) {
+			} else if (qry->flags.DNSSEC_WANT) {
 				knot_edns_set_do(pkt->opt_rr);
 				knot_wire_set_cd(pkt->wire);
 			}
@@ -712,8 +723,11 @@ int kr_resolve_begin(struct kr_request *request, struct kr_context *ctx, knot_pk
 	array_init(request->additional);
 	array_init(request->answ_selected);
 	array_init(request->auth_selected);
+	array_init(request->add_selected);
 	request->answ_validated = false;
 	request->auth_validated = false;
+	request->trace_log = NULL;
+	request->trace_finish = NULL;
 
 	/* Expect first query */
 	kr_rplan_init(&request->rplan, request, &request->pool);
@@ -742,11 +756,11 @@ static int resolve_query(struct kr_request *request, const knot_pkt_t *packet)
 	}
 
 	/* Deferred zone cut lookup for this query. */
-	qry->flags |= QUERY_AWAIT_CUT;
+	qry->flags.AWAIT_CUT = true;
 	/* Want DNSSEC if it's posible to secure this name (e.g. is covered by any TA) */
 	if ((knot_wire_get_ad(packet->wire) || knot_pkt_has_dnssec(packet)) &&
 	    kr_ta_covers_qry(request->ctx, qname, qtype)) {
-		qry->flags |= QUERY_DNSSEC_WANT;
+		qry->flags.DNSSEC_WANT = true;
 	}
 
 	/* Initialize answer packet */
@@ -758,7 +772,7 @@ static int resolve_query(struct kr_request *request, const knot_pkt_t *packet)
 
 	if (cd_is_set) {
 		knot_wire_set_cd(answer->wire);
-	} else if (qry->flags & QUERY_DNSSEC_WANT) {
+	} else if (qry->flags.DNSSEC_WANT) {
 		knot_wire_set_ad(answer->wire);
 	}
 
@@ -782,15 +796,14 @@ KR_PURE static bool kr_inaddr_equal(const struct sockaddr *a, const struct socka
 static void update_nslist_rtt(struct kr_context *ctx, struct kr_query *qry, const struct sockaddr *src)
 {
 	/* Do not track in safe mode. */
-	if (qry->flags & QUERY_SAFEMODE) {
+	if (qry->flags.SAFEMODE) {
 		return;
 	}
 
 	/* Calculate total resolution time from the time the query was generated. */
-	struct timeval now;
-	gettimeofday(&now, NULL);
-	unsigned elapsed = time_diff(&qry->timestamp, &now);
-
+	uint64_t elapsed = kr_now() - qry->timestamp_mono;
+	elapsed = elapsed > UINT_MAX ? UINT_MAX : elapsed;
+ 
 	/* NSs in the preference list prior to the one who responded will be penalised
 	 * with the RETRY timer interval. This is because we know they didn't respond
 	 * for N retries, so their RTT must be at least N * RETRY.
@@ -805,10 +818,11 @@ static void update_nslist_rtt(struct kr_context *ctx, struct kr_query *qry, cons
 		/* If this address is the source of the answer, update its RTT */
 		if (kr_inaddr_equal(src, addr)) {
 			kr_nsrep_update_rtt(&qry->ns, addr, elapsed, ctx->cache_rtt, KR_NS_UPDATE);
-			WITH_VERBOSE {
+			WITH_VERBOSE(qry) {
 				char addr_str[INET6_ADDRSTRLEN];
 				inet_ntop(addr->sa_family, kr_inaddr(addr), addr_str, sizeof(addr_str));
-				VERBOSE_MSG(qry, "<= server: '%s' rtt: %ld ms\n", addr_str, elapsed);
+				VERBOSE_MSG(qry, "<= server: '%s' rtt: %"PRIu64" ms\n",
+						addr_str, elapsed);
 			}
 		} else {
 			/* Response didn't come from this IP, but we know the RTT must be at least
@@ -817,10 +831,11 @@ static void update_nslist_rtt(struct kr_context *ctx, struct kr_query *qry, cons
 			 * that 'b' didn't respond for at least 350 - (1 * 300) ms. We can't say that
 			 * its RTT is 50ms, but we can say that its score shouldn't be less than 50. */
 			 kr_nsrep_update_rtt(&qry->ns, addr, elapsed, ctx->cache_rtt, KR_NS_MAX);
-			 WITH_VERBOSE {
+			 WITH_VERBOSE(qry) {
 			 	char addr_str[INET6_ADDRSTRLEN];
 			 	inet_ntop(addr->sa_family, kr_inaddr(addr), addr_str, sizeof(addr_str));
-			 	VERBOSE_MSG(qry, "<= server: '%s' rtt: >=%ld ms\n", addr_str, elapsed);
+				VERBOSE_MSG(qry, "<= server: '%s' rtt: >= %"PRIu64" ms\n",
+						addr_str, elapsed);
 			 }
 		}
 		/* Subtract query start time from elapsed time */
@@ -834,21 +849,22 @@ static void update_nslist_rtt(struct kr_context *ctx, struct kr_query *qry, cons
 static void update_nslist_score(struct kr_request *request, struct kr_query *qry, const struct sockaddr *src, knot_pkt_t *packet)
 {
 	struct kr_context *ctx = request->ctx;
-	/* On sucessful answer, update preference list RTT and penalise timer  */
+	/* On successful answer, update preference list RTT and penalise timer  */
 	if (request->state != KR_STATE_FAIL) {
 		/* Update RTT information for preference list */
 		update_nslist_rtt(ctx, qry, src);
 		/* Do not complete NS address resolution on soft-fail. */
 		const int rcode = packet ? knot_wire_get_rcode(packet->wire) : 0;
 		if (rcode != KNOT_RCODE_SERVFAIL && rcode != KNOT_RCODE_REFUSED) {
-			qry->flags &= ~(QUERY_AWAIT_IPV6|QUERY_AWAIT_IPV4);
+			qry->flags.AWAIT_IPV6 = false;
+			qry->flags.AWAIT_IPV4 = false;
 		} else { /* Penalize SERVFAILs. */
 			kr_nsrep_update_rtt(&qry->ns, src, KR_NS_PENALTY, ctx->cache_rtt, KR_NS_ADD);
 		}
 	/* Penalise resolution failures except validation failures. */
-	} else if (!(qry->flags & QUERY_DNSSEC_BOGUS)) {
+	} else if (!(qry->flags.DNSSEC_BOGUS)) {
 		kr_nsrep_update_rtt(&qry->ns, src, KR_NS_TIMEOUT, ctx->cache_rtt, KR_NS_RESET);
-		WITH_VERBOSE {
+		WITH_VERBOSE(qry) {
 			char addr_str[INET6_ADDRSTRLEN];
 			inet_ntop(src->sa_family, kr_inaddr(src), addr_str, sizeof(addr_str));
 			VERBOSE_MSG(qry, "=> server: '%s' flagged as 'bad'\n", addr_str);
@@ -856,6 +872,18 @@ static void update_nslist_score(struct kr_request *request, struct kr_query *qry
 	}
 }
 
+static bool resolution_time_exceeded(struct kr_query *qry, uint64_t now)
+{
+	uint64_t resolving_time = now - qry->creation_time_mono;
+	if (resolving_time > KR_RESOLVE_TIME_LIMIT) {
+		WITH_VERBOSE(qry) {
+			VERBOSE_MSG(qry, "query resolution time limit exceeded\n");
+		}
+		return true;
+	}
+	return false;
+}
+
 int kr_resolve_consume(struct kr_request *request, const struct sockaddr *src, knot_pkt_t *packet)
 {
 	struct kr_rplan *rplan = &request->rplan;
@@ -870,12 +898,16 @@ int kr_resolve_consume(struct kr_request *request, const struct sockaddr *src, k
 
 	/* Different processing for network error */
 	struct kr_query *qry = array_tail(rplan->pending);
-	bool tried_tcp = (qry->flags & QUERY_TCP);
+	/* Check overall resolution time */
+	if (resolution_time_exceeded(qry, kr_now())) {
+		return KR_STATE_FAIL;
+	}
+	bool tried_tcp = (qry->flags.TCP);
 	if (!packet || packet->size == 0) {
 		if (tried_tcp) {
 			request->state = KR_STATE_FAIL;
 		} else {
-			qry->flags |= QUERY_TCP;
+			qry->flags.TCP = true;
 		}
 	} else {
 		/* Packet cleared, derandomize QNAME. */
@@ -884,13 +916,11 @@ int kr_resolve_consume(struct kr_request *request, const struct sockaddr *src, k
 			randomized_qname_case(qname_raw, qry->secret);
 		}
 		request->state = KR_STATE_CONSUME;
-		if (qry->flags & QUERY_CACHED) {
+		if (qry->flags.CACHED) {
 			ITERATE_LAYERS(request, qry, consume, packet);
 		} else {
-			struct timeval now;
-			gettimeofday(&now, NULL);
 			/* Fill in source and latency information. */
-			request->upstream.rtt = time_diff(&qry->timestamp, &now);
+			request->upstream.rtt = kr_now() - qry->timestamp_mono;
 			request->upstream.addr = src;
 			ITERATE_LAYERS(request, qry, consume, packet);
 			/* Clear temporary information */
@@ -900,30 +930,33 @@ int kr_resolve_consume(struct kr_request *request, const struct sockaddr *src, k
 	}
 
 	/* Track RTT for iterative answers */
-	if (src && !(qry->flags & QUERY_CACHED)) {
+	if (src && !(qry->flags.CACHED)) {
 		update_nslist_score(request, qry, src, packet);
 	}
 	/* Resolution failed, invalidate current NS. */
 	if (request->state == KR_STATE_FAIL) {
 		invalidate_ns(rplan, qry);
-		qry->flags &= ~QUERY_RESOLVED;
+		qry->flags.RESOLVED = false;
 	}
 
 	/* Pop query if resolved. */
 	if (request->state == KR_STATE_YIELD) {
 		return KR_STATE_PRODUCE; /* Requery */
-	} else if (qry->flags & QUERY_RESOLVED) {
+	} else if (qry->flags.RESOLVED) {
 		kr_rplan_pop(rplan, qry);
-	} else if (!tried_tcp && (qry->flags & QUERY_TCP)) {
+	} else if (!tried_tcp && (qry->flags.TCP)) {
 		return KR_STATE_PRODUCE; /* Requery over TCP */
 	} else { /* Clear query flags for next attempt */
-		qry->flags &= ~(QUERY_CACHED|QUERY_TCP);
+		qry->flags.CACHED = false;
+		if (!request->options.TCP) {
+			qry->flags.TCP = false;
+		}
 	}
 
 	ITERATE_LAYERS(request, qry, reset);
 
 	/* Do not finish with bogus answer. */
-	if (qry->flags & QUERY_DNSSEC_BOGUS)  {
+	if (qry->flags.DNSSEC_BOGUS)  {
 		return KR_STATE_FAIL;
 	}
 
@@ -943,9 +976,9 @@ static struct kr_query *zone_cut_subreq(struct kr_rplan *rplan, struct kr_query
 	    kr_zonecut_copy_trust(&next->zone_cut, &parent->zone_cut) != 0) {
 		return NULL;
 	}
-	next->flags |= QUERY_NO_MINIMIZE;
-	if (parent->flags & QUERY_DNSSEC_WANT) {
-		next->flags |= QUERY_DNSSEC_WANT;
+	next->flags.NO_MINIMIZE = true;
+	if (parent->flags.DNSSEC_WANT) {
+		next->flags.DNSSEC_WANT = true;
 	}
 	return next;
 }
@@ -957,40 +990,40 @@ static int forward_trust_chain_check(struct kr_request *request, struct kr_query
 	map_t *negative_anchors = &request->ctx->negative_anchors;
 
 	if (qry->parent != NULL &&
-	    !(qry->forward_flags & QUERY_CNAME) &&
-	    !(qry->flags & QUERY_DNS64_MARK) &&
+	    !(qry->forward_flags.CNAME) &&
+	    !(qry->flags.DNS64_MARK) &&
 	    knot_dname_in(qry->parent->zone_cut.name, qry->zone_cut.name)) {
 		return KR_STATE_PRODUCE;
 	}
 
-	assert(qry->flags & QUERY_FORWARD);
+	assert(qry->flags.FORWARD);
 
 	if (!trust_anchors) {
-		qry->flags &= ~QUERY_AWAIT_CUT;
+		qry->flags.AWAIT_CUT = false;
 		return KR_STATE_PRODUCE;
 	}
 
-	if (qry->flags & QUERY_DNSSEC_INSECURE) {
-		qry->flags &= ~QUERY_AWAIT_CUT;
+	if (qry->flags.DNSSEC_INSECURE) {
+		qry->flags.AWAIT_CUT = false;
 		return KR_STATE_PRODUCE;
 	}
 
-	if (qry->forward_flags & QUERY_NO_MINIMIZE) {
-		qry->flags &= ~QUERY_AWAIT_CUT;
+	if (qry->forward_flags.NO_MINIMIZE) {
+		qry->flags.AWAIT_CUT = false;
 		return KR_STATE_PRODUCE;
 	}
 
 	const knot_dname_t *wanted_name = qry->sname;
 	const knot_dname_t *start_name = qry->sname;
-	if ((qry->flags & QUERY_AWAIT_CUT) && !resume) {
-		qry->flags &= ~QUERY_AWAIT_CUT;
+	if ((qry->flags.AWAIT_CUT) && !resume) {
+		qry->flags.AWAIT_CUT = false;
 		const knot_dname_t *longest_ta = kr_ta_get_longest_name(trust_anchors, qry->sname);
 		if (longest_ta) {
 			start_name = longest_ta;
 			qry->zone_cut.name = knot_dname_copy(start_name, qry->zone_cut.pool);
-			qry->flags |= QUERY_DNSSEC_WANT;
+			qry->flags.DNSSEC_WANT = true;
 		} else {
-			qry->flags &= ~QUERY_DNSSEC_WANT;
+			qry->flags.DNSSEC_WANT = false;
 			return KR_STATE_PRODUCE;
 		}
 	}
@@ -1038,19 +1071,19 @@ static int forward_trust_chain_check(struct kr_request *request, struct kr_query
 			    knot_dname_is_equal(q->sname, wanted_name)) {
 				if (q->stype == KNOT_RRTYPE_DS) {
 					ds_req = true;
-					if (q->flags & QUERY_DNSSEC_NODS) {
+					if (q->flags.DNSSEC_NODS) {
 						nods = true;
 					}
-					if (q->flags & QUERY_CNAME) {
+					if (q->flags.CNAME) {
 						nods = true;
 						ns_exist = false;
-					} else if (!(q->flags & QUERY_DNSSEC_OPTOUT)) {
+					} else if (!(q->flags.DNSSEC_OPTOUT)) {
 						int ret = kr_dnssec_matches_name_and_type(&request->auth_selected, q->uid,
 											  wanted_name, KNOT_RRTYPE_NS);
 						ns_exist = (ret == kr_ok());
 					}
 				} else {
-					if (q->flags & QUERY_CNAME) {
+					if (q->flags.CNAME) {
 						nods = true;
 						ns_exist = false;
 					}
@@ -1068,7 +1101,7 @@ static int forward_trust_chain_check(struct kr_request *request, struct kr_query
 			return KR_STATE_DONE;
 		}
 
-		if (qry->parent == NULL && (qry->flags & QUERY_CNAME) &&
+		if (qry->parent == NULL && (qry->flags.CNAME) &&
 		    ds_req && ns_req) {
 			return KR_STATE_PRODUCE;
 		}
@@ -1089,18 +1122,18 @@ static int forward_trust_chain_check(struct kr_request *request, struct kr_query
 	/* Disable DNSSEC if it enters NTA. */
 	if (kr_ta_get(negative_anchors, wanted_name)){
 		VERBOSE_MSG(qry, ">< negative TA, going insecure\n");
-		qry->flags &= ~QUERY_DNSSEC_WANT;
+		qry->flags.DNSSEC_WANT = false;
 	}
 
 	/* Enable DNSSEC if enters a new island of trust. */
-	bool want_secured = (qry->flags & QUERY_DNSSEC_WANT) &&
+	bool want_secured = (qry->flags.DNSSEC_WANT) &&
 			    !knot_wire_get_cd(request->answer->wire);
-	if (!(qry->flags & QUERY_DNSSEC_WANT) &&
+	if (!(qry->flags.DNSSEC_WANT) &&
 	    !knot_wire_get_cd(request->answer->wire) &&
 	    kr_ta_get(trust_anchors, wanted_name)) {
-		qry->flags |= QUERY_DNSSEC_WANT;
+		qry->flags.DNSSEC_WANT = true;
 		want_secured = true;
-		WITH_VERBOSE {
+		WITH_VERBOSE(qry) {
 		char qname_str[KNOT_DNAME_MAXLEN];
 		knot_dname_to_str(qname_str, wanted_name, sizeof(qname_str));
 		VERBOSE_MSG(qry, ">< TA: '%s'\n", qname_str);
@@ -1155,23 +1188,23 @@ static int trust_chain_check(struct kr_request *request, struct kr_query *qry)
 	/* Disable DNSSEC if it enters NTA. */
 	if (kr_ta_get(negative_anchors, qry->zone_cut.name)){
 		VERBOSE_MSG(qry, ">< negative TA, going insecure\n");
-		qry->flags &= ~QUERY_DNSSEC_WANT;
-		qry->flags |= QUERY_DNSSEC_INSECURE;
+		qry->flags.DNSSEC_WANT = false;
+		qry->flags.DNSSEC_INSECURE = true;
 	}
-	if (qry->flags & QUERY_DNSSEC_NODS) {
+	if (qry->flags.DNSSEC_NODS) {
 		/* This is the next query iteration with minimized qname.
 		 * At previous iteration DS non-existance has been proven */
-		qry->flags &= ~QUERY_DNSSEC_NODS;
-		qry->flags &= ~QUERY_DNSSEC_WANT;
-		qry->flags |= QUERY_DNSSEC_INSECURE;
+		qry->flags.DNSSEC_NODS = false;
+		qry->flags.DNSSEC_WANT = false;
+		qry->flags.DNSSEC_INSECURE = true;
 	}
 	/* Enable DNSSEC if entering a new (or different) island of trust,
 	 * and update the TA RRset if required. */
-	bool want_secured = (qry->flags & QUERY_DNSSEC_WANT) &&
+	bool want_secured = (qry->flags.DNSSEC_WANT) &&
 			    !knot_wire_get_cd(request->answer->wire);
 	knot_rrset_t *ta_rr = kr_ta_get(trust_anchors, qry->zone_cut.name);
 	if (!knot_wire_get_cd(request->answer->wire) && ta_rr) {
-		qry->flags |= QUERY_DNSSEC_WANT;
+		qry->flags.DNSSEC_WANT = true;
 		want_secured = true;
 
 		if (qry->zone_cut.trust_anchor == NULL
@@ -1179,7 +1212,7 @@ static int trust_chain_check(struct kr_request *request, struct kr_query *qry)
 			mm_free(qry->zone_cut.pool, qry->zone_cut.trust_anchor);
 			qry->zone_cut.trust_anchor = knot_rrset_copy(ta_rr, qry->zone_cut.pool);
 
-			WITH_VERBOSE {
+			WITH_VERBOSE(qry) {
 			char qname_str[KNOT_DNAME_MAXLEN];
 			knot_dname_to_str(qname_str, ta_rr->owner, sizeof(qname_str));
 			VERBOSE_MSG(qry, ">< TA: '%s'\n", qname_str);
@@ -1197,7 +1230,8 @@ static int trust_chain_check(struct kr_request *request, struct kr_query *qry)
 		if (!next) {
 			return KR_STATE_FAIL;
 		}
-		next->flags |= (QUERY_AWAIT_CUT | QUERY_DNSSEC_WANT);
+		next->flags.AWAIT_CUT = true;
+		next->flags.DNSSEC_WANT = true;
 		return KR_STATE_DONE;
 	}
 	/* Try to fetch missing DNSKEY (either missing or above current cut).
@@ -1217,19 +1251,20 @@ static int trust_chain_check(struct kr_request *request, struct kr_query *qry)
 
 /** @internal Check current zone cut status and credibility, spawn subrequests if needed. */
 static int zone_cut_check(struct kr_request *request, struct kr_query *qry, knot_pkt_t *packet)
+/* TODO: using cache on this point in this way just isn't nice; remove in time */
 {
 	/* Stub mode, just forward and do not solve cut. */
-	if (qry->flags & QUERY_STUB) {
+	if (qry->flags.STUB) {
 		return KR_STATE_PRODUCE;
 	}
 
 	/* Forwarding to upstream resolver mode.
 	 * Since forwarding targets already are in qry->ns -
 	 * cut fetching is not needed. */
-	if (qry->flags & QUERY_FORWARD) {
+	if (qry->flags.FORWARD) {
 		return forward_trust_chain_check(request, qry, false);
 	}
-	if (!(qry->flags & QUERY_AWAIT_CUT)) {
+	if (!(qry->flags.AWAIT_CUT)) {
 		/* The query was resolved from cache.
 		 * Spawn DS \ DNSKEY requests if needed and exit */
 		return trust_chain_check(request, qry);
@@ -1243,8 +1278,8 @@ static int zone_cut_check(struct kr_request *request, struct kr_query *qry, knot
 		if (ret != 0) {
 			return KR_STATE_FAIL;
 		}
-		VERBOSE_MSG(qry, "=> using root hints\n");
-		qry->flags &= ~QUERY_AWAIT_CUT;
+		VERBOSE_MSG(qry, "=> no cache open, using root hints\n");
+		qry->flags.AWAIT_CUT = false;
 		return KR_STATE_DONE;
 	}
 
@@ -1273,12 +1308,12 @@ static int zone_cut_check(struct kr_request *request, struct kr_query *qry, knot
 	} while (state == KR_STATE_CONSUME);
 
 	/* Update minimized QNAME if zone cut changed */
-	if (qry->zone_cut.name[0] != '\0' && !(qry->flags & QUERY_NO_MINIMIZE)) {
+	if (qry->zone_cut.name && qry->zone_cut.name[0] != '\0' && !(qry->flags.NO_MINIMIZE)) {
 		if (kr_make_query(qry, packet) != 0) {
 			return KR_STATE_FAIL;
 		}
 	}
-	qry->flags &= ~QUERY_AWAIT_CUT;
+	qry->flags.AWAIT_CUT = false;
 
 	/* Check trust chain */
 	return trust_chain_check(request, qry);
@@ -1298,8 +1333,8 @@ int kr_resolve_produce(struct kr_request *request, struct sockaddr **dst, int *t
 	if (qry->deferred != NULL) {
 		/* @todo: Refactoring validator, check trust chain before resuming. */
 		int state = 0;
-		if (((qry->flags & QUERY_FORWARD) == 0) ||
-		    ((qry->stype == KNOT_RRTYPE_DS) && (qry->flags & QUERY_CNAME))) {
+		if (((qry->flags.FORWARD) == 0) ||
+		    ((qry->stype == KNOT_RRTYPE_DS) && (qry->flags.CNAME))) {
 			state = trust_chain_check(request, qry);
 		} else {
 			state = forward_trust_chain_check(request, qry, true);
@@ -1323,7 +1358,7 @@ int kr_resolve_produce(struct kr_request *request, struct sockaddr **dst, int *t
 	} else {
 		/* Caller is interested in always tracking a zone cut, even if the answer is cached
 		 * this is normally not required, and incurrs another cache lookups for cached answer. */
-		if (qry->flags & QUERY_ALWAYS_CUT) {
+		if (qry->flags.ALWAYS_CUT) {
 			switch(zone_cut_check(request, qry, packet)) {
 			case KR_STATE_FAIL: return KR_STATE_FAIL;
 			case KR_STATE_DONE: return KR_STATE_PRODUCE;
@@ -1345,7 +1380,7 @@ int kr_resolve_produce(struct kr_request *request, struct sockaddr **dst, int *t
 	case KR_STATE_CONSUME: break;
 	case KR_STATE_DONE:
 	default: /* Current query is done */
-		if (qry->flags & QUERY_RESOLVED && request->state != KR_STATE_YIELD) {
+		if (qry->flags.RESOLVED && request->state != KR_STATE_YIELD) {
 			kr_rplan_pop(rplan, qry);
 		}
 		ITERATE_LAYERS(request, qry, reset);
@@ -1360,7 +1395,7 @@ int kr_resolve_produce(struct kr_request *request, struct sockaddr **dst, int *t
 	}
 
 	/* Update zone cut, spawn new subrequests. */
-	if (!(qry->flags & QUERY_STUB)) {
+	if (!(qry->flags.STUB)) {
 		int state = zone_cut_check(request, qry, packet);
 		switch(state) {
 		case KR_STATE_FAIL: return KR_STATE_FAIL;
@@ -1379,20 +1414,25 @@ ns_election:
 		return KR_STATE_FAIL;
 	}
 
-	const bool retry = (qry->flags & (QUERY_TCP|QUERY_BADCOOKIE_AGAIN));
-	if (qry->flags & (QUERY_AWAIT_IPV4|QUERY_AWAIT_IPV6)) {
+	const struct kr_qflags qflg = qry->flags;
+	const bool retry = qflg.TCP || qflg.BADCOOKIE_AGAIN;
+	if (qflg.AWAIT_IPV4 || qflg.AWAIT_IPV6) {
 		kr_nsrep_elect_addr(qry, request->ctx);
-	} else if (qry->flags & (QUERY_FORWARD|QUERY_STUB)) {
+	} else if (qflg.FORWARD || qflg.STUB) {
 		kr_nsrep_sort(&qry->ns, request->ctx->cache_rtt);
 	} else if (!qry->ns.name || !retry) { /* Keep NS when requerying/stub/badcookie. */
 		/* Root DNSKEY must be fetched from the hints to avoid chicken and egg problem. */
 		if (qry->sname[0] == '\0' && qry->stype == KNOT_RRTYPE_DNSKEY) {
 			kr_zonecut_set_sbelt(request->ctx, &qry->zone_cut);
-			qry->flags |= QUERY_NO_THROTTLE; /* Pick even bad SBELT servers */
+			qry->flags.NO_THROTTLE = true; /* Pick even bad SBELT servers */
 		}
 		kr_nsrep_elect(qry, request->ctx);
 		if (qry->ns.score > KR_NS_MAX_SCORE) {
-			VERBOSE_MSG(qry, "=> no valid NS left\n");
+			if (!qry->zone_cut.nsset.root) {
+				VERBOSE_MSG(qry, "=> no NS with an address\n");
+			} else {
+				VERBOSE_MSG(qry, "=> no valid NS left\n");
+			}
 			ITERATE_LAYERS(request, qry, reset);
 			kr_rplan_pop(rplan, qry);
 			return KR_STATE_PRODUCE;
@@ -1403,7 +1443,9 @@ ns_election:
 	if (qry->ns.addr[0].ip.sa_family == AF_UNSPEC) {
 		int ret = ns_resolve_addr(qry, request);
 		if (ret != 0) {
-			qry->flags &= ~(QUERY_AWAIT_IPV6|QUERY_AWAIT_IPV4|QUERY_TCP);
+			qry->flags.AWAIT_IPV6 = false;
+			qry->flags.AWAIT_IPV4 = false;
+			qry->flags.TCP = false;
 			qry->ns.name = NULL;
 			goto ns_election; /* Must try different NS */
 		}
@@ -1412,7 +1454,7 @@ ns_election:
 	}
 
 	/* Randomize query case (if not in safemode or turned off) */
-	qry->secret = (qry->flags & (QUERY_SAFEMODE | QUERY_NO_0X20))
+	qry->secret = (qry->flags.SAFEMODE || qry->flags.NO_0X20)
 			? 0 : kr_rand_uint(0);
 	knot_dname_t *qname_raw = (knot_dname_t *)knot_pkt_qname(packet);
 	randomized_qname_case(qname_raw, qry->secret);
@@ -1421,10 +1463,9 @@ ns_election:
 	 * Additional query is going to be finalised when calling
 	 * kr_resolve_checkout().
 	 */
-
-	gettimeofday(&qry->timestamp, NULL);
+	qry->timestamp_mono = kr_now();
 	*dst = &qry->ns.addr[0].ip;
-	*type = (qry->flags & QUERY_TCP) ? SOCK_STREAM : SOCK_DGRAM;
+	*type = (qry->flags.TCP) ? SOCK_STREAM : SOCK_DGRAM;
 	return request->state;
 }
 
@@ -1496,7 +1537,7 @@ int kr_resolve_checkout(struct kr_request *request, struct sockaddr *src,
 		return kr_error(EINVAL);
 	}
 
-	WITH_VERBOSE {
+	WITH_VERBOSE(qry) {
 	char qname_str[KNOT_DNAME_MAXLEN], zonecut_str[KNOT_DNAME_MAXLEN], ns_str[INET6_ADDRSTRLEN], type_str[16];
 	knot_dname_to_str(qname_str, knot_pkt_qname(packet), sizeof(qname_str));
 	knot_dname_to_str(zonecut_str, qry->zone_cut.name, sizeof(zonecut_str));
@@ -1511,7 +1552,7 @@ int kr_resolve_checkout(struct kr_request *request, struct sockaddr *src,
 		}
 		inet_ntop(addr->sa_family, kr_inaddr(&qry->ns.addr[i].ip), ns_str, sizeof(ns_str));
 		VERBOSE_MSG(qry, "=> querying: '%s' score: %u zone cut: '%s' m12n: '%s' type: '%s' proto: '%s'\n",
-			ns_str, qry->ns.score, zonecut_str, qname_str, type_str, (qry->flags & QUERY_TCP) ? "tcp" : "udp");
+			ns_str, qry->ns.score, zonecut_str, qname_str, type_str, (qry->flags.TCP) ? "tcp" : "udp");
 		break;
 	}}
 
@@ -1537,8 +1578,20 @@ int kr_resolve_finish(struct kr_request *request, int state)
 
 	request->state = state;
 	ITERATE_LAYERS(request, NULL, finish);
-	VERBOSE_MSG(NULL, "finished: %d, queries: %zu, mempool: %zu B\n",
+
+	struct kr_query *last = kr_rplan_last(rplan);
+	VERBOSE_MSG(last, "finished: %d, queries: %zu, mempool: %zu B\n",
 	          request->state, rplan->resolved.len, (size_t) mp_total_size(request->pool.ctx));
+
+	/* Trace request finish */
+	if (request->trace_finish) {
+		request->trace_finish(request);
+	}
+
+	/* Uninstall all tracepoints */
+	request->trace_finish = NULL;
+	request->trace_log = NULL;
+
 	return KR_STATE_DONE;
 }
 
