@@ -19,7 +19,7 @@
 
 #include "lib/rplan.h"
 #include "lib/resolve.h"
-#include "lib/cache.h"
+#include "lib/cache/api.h"
 #include "lib/defines.h"
 #include "lib/layer.h"
 
@@ -27,17 +27,58 @@
 #define QUERY_PROVIDES(q, name, cls, type) \
     ((q)->sclass == (cls) && (q)->stype == type && knot_dname_is_equal((q)->sname, name))
 
-/** @internal LUT of query flag names. */
-const knot_lookup_t query_flag_names[] = {
-	#define X(flag, _) { QUERY_ ## flag, #flag },
-	QUERY_FLAGS(X)
-	#undef X
-	{ 0, NULL }
-};
-
-const knot_lookup_t *kr_query_flag_names(void)
+inline static unsigned char chars_or(const unsigned char a, const unsigned char b)
 {
-	return query_flag_names;
+	return a | b;
+}
+
+/** Bits set to 1 in variable b will be set to zero in variable a. */
+inline static unsigned char chars_mask(const unsigned char a, const unsigned char b)
+{
+	return a & ~b;
+}
+
+/** Apply mod(a, b) to every byte a, b from fl1, fl2 and return result in fl1. */
+inline static void kr_qflags_mod(struct kr_qflags *fl1, struct kr_qflags fl2,
+			unsigned char mod(const unsigned char a, const unsigned char b))
+{
+	if (!fl1) abort();
+	union {
+		struct kr_qflags flags;
+		/* C99 section 6.5.3.4: sizeof(char) == 1 */
+		unsigned char chars[sizeof(struct kr_qflags)];
+	} tmp1, tmp2;
+	/* The compiler should be able to optimize all this into simple ORs. */
+	tmp1.flags = *fl1;
+	tmp2.flags = fl2;
+	for (size_t i = 0; i < sizeof(struct kr_qflags); ++i) {
+		tmp1.chars[i] = mod(tmp1.chars[i], tmp2.chars[i]);
+	}
+	*fl1 = tmp1.flags;
+}
+
+/**
+ * Set bits from variable fl2 in variable fl1.
+ * Bits which are not set in fl2 are not modified in fl1.
+ *
+ * @param[in,out] fl1
+ * @param[in] fl2
+ */
+void kr_qflags_set(struct kr_qflags *fl1, struct kr_qflags fl2)
+{
+	kr_qflags_mod(fl1, fl2, chars_or);
+}
+
+/**
+ * Clear bits from variable fl2 in variable fl1.
+ * Bits which are not set in fl2 are not modified in fl1.
+ *
+ * @param[in,out] fl1
+ * @param[in] fl2
+ */
+void kr_qflags_clear(struct kr_qflags *fl1, struct kr_qflags fl2)
+{
+	kr_qflags_mod(fl1, fl2, chars_mask);
 }
 
 static struct kr_query *query_create(knot_mm_t *pool, const knot_dname_t *name, uint32_t uid)
@@ -131,16 +172,19 @@ static struct kr_query *kr_rplan_push_query(struct kr_rplan *rplan,
 	/* Class and type must be set outside this function. */
 	qry->flags = rplan->request->options;
 	qry->parent = parent;
+	qry->request = rplan->request;
 	qry->ns.ctx = rplan->request->ctx;
 	qry->ns.addr[0].ip.sa_family = AF_UNSPEC;
 	gettimeofday(&qry->timestamp, NULL);
+	qry->timestamp_mono = kr_now();
+	qry->creation_time_mono = parent ? parent->creation_time_mono : qry->timestamp_mono;
 	kr_zonecut_init(&qry->zone_cut, (const uint8_t *)"", rplan->pool);
-	qry->reorder = qry->flags & QUERY_REORDER_RR
+	qry->reorder = qry->flags.REORDER_RR
 		? knot_wire_get_id(rplan->request->answer->wire)
 		: 0;
 
 	/* When forwarding, keep the nameserver addresses. */
-	if (parent && (parent->flags & qry->flags & QUERY_FORWARD)) {
+	if (parent && parent->flags.FORWARD && qry->flags.FORWARD) {
 		ret = kr_nsrep_copy_set(&qry->ns, &parent->ns);
 		if (ret) {
 			query_free(rplan->pool, qry);
@@ -164,8 +208,8 @@ struct kr_query *kr_rplan_push_empty(struct kr_rplan *rplan, struct kr_query *pa
 		return NULL;
 	}
 
-	WITH_VERBOSE {
-	VERBOSE_MSG(parent, "plan '%s' type '%s'\n", "", "");
+	WITH_VERBOSE(qry) {
+	VERBOSE_MSG(qry, "plan '%s' type '%s'\n", "", "");
 	}
 	return qry;
 }
@@ -185,7 +229,7 @@ struct kr_query *kr_rplan_push(struct kr_rplan *rplan, struct kr_query *parent,
 	qry->sclass = cls;
 	qry->stype = type;
 
-	WITH_VERBOSE {
+	WITH_VERBOSE(qry) {
 	char name_str[KNOT_DNAME_MAXLEN], type_str[16];
 	knot_dname_to_str(name_str, name, sizeof(name_str));
 	knot_rrtype_to_string(type, type_str, sizeof(type_str));
@@ -207,9 +251,9 @@ int kr_rplan_pop(struct kr_rplan *rplan, struct kr_query *qry)
 	}
 
 	/* Find the query, it will likely be on top */
-	for (size_t i = rplan->pending.len; i --> 0;) {
-		if (rplan->pending.at[i] == qry) {
-			array_del(rplan->pending, i);
+	for (size_t i = rplan->pending.len; i > 0; i--) {
+		if (rplan->pending.at[i - 1] == qry) {
+			array_del(rplan->pending, i - 1);
 			array_push(rplan->resolved, qry);
 			break;
 		}
@@ -236,6 +280,15 @@ struct kr_query *kr_rplan_resolved(struct kr_rplan *rplan)
 	return array_tail(rplan->resolved);
 }
 
+struct kr_query *kr_rplan_last(struct kr_rplan *rplan)
+{
+	if (!kr_rplan_empty(rplan)) {
+		return array_tail(rplan->pending);
+	}
+
+	return kr_rplan_resolved(rplan);
+}
+
 struct kr_query *kr_rplan_find_resolved(struct kr_rplan *rplan, struct kr_query *parent,
                                const knot_dname_t *name, uint16_t cls, uint16_t type)
 {
