@@ -20,78 +20,93 @@
 #include <libknot/dname.h>
 #include <libknot/codes.h>
 
-#include "lib/cache.h"
+#include "lib/cache/api.h"
 #include "lib/zonecut.h"
 #include "lib/nsrep.h"
 
-#define QUERY_FLAGS(X) \
-	X(NO_MINIMIZE,     1 << 0) /**< Don't minimize QNAME. */ \
-	X(NO_THROTTLE,     1 << 1) /**< No query/slow NS throttling. */ \
-	X(NO_IPV6,         1 << 2) /**< Disable IPv6 */ \
-	X(NO_IPV4,         1 << 3) /**< Disable IPv4 */ \
-	X(TCP,             1 << 4) /**< Use TCP for this query. */ \
-	X(RESOLVED,        1 << 5) /**< Query is resolved. */ \
-	X(AWAIT_IPV4,      1 << 6) /**< Query is waiting for A address. */ \
-	X(AWAIT_IPV6,      1 << 7) /**< Query is waiting for AAAA address. */ \
-	X(AWAIT_CUT,       1 << 8) /**< Query is waiting for zone cut lookup */ \
-	X(SAFEMODE,        1 << 9) /**< Don't use fancy stuff (EDNS, 0x20, ...) */ \
-	X(CACHED,          1 << 10) /**< Query response is cached. */ \
-	X(NO_CACHE,        1 << 11) /**< No cache for lookup; exception: finding NSs and subqueries. */ \
-	X(EXPIRING,        1 << 12) /**< Query response is cached, but expiring. */ \
-	X(ALLOW_LOCAL,     1 << 13) /**< Allow queries to local or private address ranges. */ \
-	X(DNSSEC_WANT,     1 << 14) /**< Want DNSSEC secured answer; exception: +cd, \
-				     *   i.e. knot_wire_set_cd(request->answer->wire). */ \
-	X(DNSSEC_BOGUS,    1 << 15) /**< Query response is DNSSEC bogus. */ \
-	X(DNSSEC_INSECURE, 1 << 16) /**< Query response is DNSSEC insecure. */ \
-	X(STUB,            1 << 17) /**< Stub resolution, accept received answer as solved. */ \
-	X(ALWAYS_CUT,      1 << 18) /**< Always recover zone cut (even if cached). */ \
-	X(DNSSEC_WEXPAND,  1 << 19) /**< Query response has wildcard expansion. */ \
-	X(PERMISSIVE,      1 << 20) /**< Permissive resolver mode. */ \
-	X(STRICT,          1 << 21) /**< Strict resolver mode. */ \
-	X(BADCOOKIE_AGAIN, 1 << 22) /**< Query again because bad cookie returned. */ \
-	X(CNAME,           1 << 23) /**< Query response contains CNAME in answer section. */ \
-	X(REORDER_RR,      1 << 24) /**< Reorder cached RRs. */ \
-	X(TRACE,	   1 << 25) /**< Log answer with kr_verbose_log(), unless -DNDEBUG. */ \
-	X(NO_0X20,	   1 << 26) /**< Disable query case randomization . */ \
-	X(DNSSEC_NODS,	   1 << 27) /**< DS non-existance is proven */ \
-	X(DNSSEC_OPTOUT,   1 << 28) /**< Closest encloser proof has optout */ \
-	X(NONAUTH,         1 << 29) /**< Non-authoritative in-bailiwick records are enough.
-				     *   TODO: utilize this also outside cache. */ \
-	X(FORWARD,	   1 << 30) /**< Forward all queries to upstream; validate answers */ \
-	X(DNS64_MARK,	   1u << 31) /**< Internal to ../modules/dns64/dns64.lua */
-
 /** Query flags */
-enum kr_query_flag {
-	#define X(flag, val) QUERY_ ## flag = val,
-	QUERY_FLAGS(X)
-	#undef X
+struct kr_qflags {
+	bool NO_MINIMIZE : 1;    /**< Don't minimize QNAME. */
+	bool NO_THROTTLE : 1;    /**< No query/slow NS throttling. */
+	bool NO_IPV6 : 1;        /**< Disable IPv6 */
+	bool NO_IPV4 : 1;        /**< Disable IPv4 */
+	bool TCP : 1;            /**< Use TCP for this query. */
+	bool RESOLVED : 1;       /**< Query is resolved. */
+	bool AWAIT_IPV4 : 1;     /**< Query is waiting for A address. */
+	bool AWAIT_IPV6 : 1;     /**< Query is waiting for AAAA address. */
+	bool AWAIT_CUT : 1;      /**< Query is waiting for zone cut lookup */
+	bool SAFEMODE : 1;       /**< Don't use fancy stuff (EDNS, 0x20, ...) */
+	bool CACHED : 1;         /**< Query response is cached. */
+	bool NO_CACHE : 1;       /**< No cache for lookup; exception: finding NSs and subqueries. */
+	bool EXPIRING : 1;       /**< Query response is cached, but expiring. */
+	bool ALLOW_LOCAL : 1;    /**< Allow queries to local or private address ranges. */
+	bool DNSSEC_WANT : 1;    /**< Want DNSSEC secured answer; exception: +cd,
+				  * i.e. knot_wire_set_cd(request->answer->wire). */
+	bool DNSSEC_BOGUS : 1;   /**< Query response is DNSSEC bogus. */
+	bool DNSSEC_INSECURE : 1;/**< Query response is DNSSEC insecure. */
+	bool DNSSEC_CD : 1;      /**< CD bit in query */
+	bool STUB : 1;           /**< Stub resolution, accept received answer as solved. */
+	bool ALWAYS_CUT : 1;     /**< Always recover zone cut (even if cached). */
+	bool DNSSEC_WEXPAND : 1; /**< Query response has wildcard expansion. */
+	bool PERMISSIVE : 1;     /**< Permissive resolver mode. */
+	bool STRICT : 1;         /**< Strict resolver mode. */
+	bool BADCOOKIE_AGAIN : 1;/**< Query again because bad cookie returned. */
+	bool CNAME : 1;          /**< Query response contains CNAME in answer section. */
+	bool REORDER_RR : 1;     /**< Reorder cached RRs. */
+	bool TRACE : 1;          /**< Log answer with kr_verbose_log(), unless -DNDEBUG. */
+	bool NO_0X20 : 1;        /**< Disable query case randomization . */
+	bool DNSSEC_NODS : 1;    /**< DS non-existance is proven */
+	bool DNSSEC_OPTOUT : 1;  /**< Closest encloser proof has optout */
+	bool NONAUTH : 1;        /**< Non-authoritative in-bailiwick records are enough.
+				  * TODO: utilize this also outside cache. */
+	bool FORWARD : 1;        /**< Forward all queries to upstream; validate answers. */
+	bool DNS64_MARK : 1;     /**< Internal mark for dns64 module. */
+	bool CACHE_TRIED : 1;    /**< Internal to cache module. */
 };
 
-/** Query flag names table */
-KR_EXPORT KR_CONST
-const knot_lookup_t *kr_query_flag_names(void);
+/** Combine flags together.  This means set union for simple flags. */
+KR_EXPORT
+void kr_qflags_set(struct kr_qflags *fl1, struct kr_qflags fl2);
+
+/** Remove flags.  This means set-theoretic difference. */
+KR_EXPORT
+void kr_qflags_clear(struct kr_qflags *fl1, struct kr_qflags fl2);
+
+/** Callback for serve-stale decisions.
+ * @param ttl the expired TTL (i.e. it's < 0)
+ * @return the adjusted TTL (typically 1) or < 0.
+ */
+typedef int32_t (*kr_stale_cb)(int32_t ttl, const knot_dname_t *owner, uint16_t type,
+				const struct kr_query *qry);
 
 /**
  * Single query representation.
  */
 struct kr_query {
 	struct kr_query *parent;
-	knot_dname_t *sname;
+	knot_dname_t *sname; /**< The name to resolve - lower-cased, uncompressed. */
 	uint16_t stype;
 	uint16_t sclass;
 	uint16_t id;
-	uint32_t flags;
+	struct kr_qflags flags, forward_flags;
 	uint32_t secret;
 	uint16_t fails;
 	uint16_t reorder; /**< Seed to reorder (cached) RRs in answer or zero. */
-	struct timeval timestamp;
+	uint64_t creation_time_mono; /* The time of query's creation (milliseconds).
+				 * Or time of creation of an oldest
+				 * ancestor if it is a subquery. */
+	uint64_t timestamp_mono; /**< Time of query created or time of
+	                           * query to upstream resolver (milliseconds). */
+	struct timeval timestamp; /**< Real time for TTL+DNSSEC checks (.tv_sec only). */
 	struct kr_zonecut zone_cut;
-	struct kr_nsrep ns;
 	struct kr_layer_pickle *deferred;
 	uint32_t uid; /**< Query iteration number, unique within the kr_rplan. */
-	uint32_t forward_flags;
 	/** Pointer to the query that originated this one because of following a CNAME (or NULL). */
 	struct kr_query *cname_parent;
+	struct kr_request *request; /**< Parent resolution request. */
+	kr_stale_cb stale_cb; /**< See the type */
+	/* Beware: this must remain the last, because of lua bindings. */
+	struct kr_nsrep ns;
 };
 
 /** @cond internal Array of queries. */
@@ -182,9 +197,13 @@ bool kr_rplan_satisfies(struct kr_query *closure, const knot_dname_t *name, uint
 KR_EXPORT KR_PURE
 struct kr_query *kr_rplan_resolved(struct kr_rplan *rplan);
 
-/** Return query predecessor. */
+/**
+  * Return last query (either currently being solved or last resolved).
+  * This is necessary to retrieve the last query in case of resolution failures (e.g. time limit reached).
+  */
 KR_EXPORT KR_PURE
-struct kr_query *kr_rplan_next(struct kr_query *qry);
+struct kr_query *kr_rplan_last(struct kr_rplan *rplan);
+
 
 /**
  * Check if a given query already resolved.
