@@ -21,7 +21,10 @@
 #include <arpa/inet.h>
 #include <sys/time.h>
 #include <contrib/cleanup.h>
+#include <contrib/ccan/asprintf/asprintf.h>
 #include <ccan/isaac/isaac.h>
+#include <ucw/mempool.h>
+#include <gnutls/gnutls.h>
 #include <libknot/descriptor.h>
 #include <libknot/dname.h>
 #include <libknot/rrtype/rrsig.h>
@@ -74,10 +77,22 @@ static inline int u16tostr(uint8_t *dst, uint16_t num)
  * Cleanup callbacks.
  */
 
+static void kres_gnutls_log(int level, const char *message)
+{
+	kr_log_verbose("[gnutls] (%d) %s", level, message);
+}
+
 bool kr_verbose_set(bool status)
 {
 #ifndef NOVERBOSELOG
 	kr_verbose_status = status;
+
+	/* gnutls logs messages related to our TLS and also libdnssec,
+	 * and the logging is set up in a global way only */
+	if (status) {
+		gnutls_global_set_log_function(kres_gnutls_log);
+	}
+	gnutls_global_set_log_level(status ? 5 : 0);
 #endif
 	return kr_verbose_status;
 }
@@ -93,21 +108,53 @@ void kr_log_verbose(const char *fmt, ...)
 	}
 }
 
+bool kr_log_trace(const struct kr_query *query, const char *source, const char *fmt, ...)
+{
+	if (!kr_log_trace_enabled(query)) {
+		return false;
+	}
+
+	auto_free char *msg = NULL;
+
+	va_list args;
+	va_start(args, fmt);
+	int len = vasprintf(&msg, fmt, args);
+	va_end(args);
+
+	/* Check formatting result before logging */
+	if (len < 0) {
+		return false;
+	}
+
+	query->request->trace_log(query, source, msg);
+	return true;
+}
+
 char* kr_strcatdup(unsigned n, ...)
 {
+	if (n < 1) {
+		return NULL;
+	}
+
 	/* Calculate total length */
 	size_t total_len = 0;
 	va_list vl;
 	va_start(vl, n);
 	for (unsigned i = 0; i < n; ++i) {
 		char *item = va_arg(vl, char *);
-		total_len += strlen_safe(item);
+		const size_t new_len = total_len + strlen_safe(item);
+		if (unlikely(new_len < total_len)) {
+			va_end(vl);
+			return NULL;
+		}
+		total_len = new_len;
 	}
 	va_end(vl);
 
 	/* Allocate result and fill */
 	char *result = NULL;
 	if (total_len > 0) {
+		if (unlikely(total_len + 1 == 0)) return NULL;
 		result = malloc(total_len + 1);
 	}
 	if (result) {
@@ -290,6 +337,18 @@ int kr_inaddr_len(const struct sockaddr *addr)
 	return kr_family_len(addr->sa_family);
 }
 
+int kr_sockaddr_len(const struct sockaddr *addr)
+{
+	if (!addr) {
+		return kr_error(EINVAL);
+	}
+	switch (addr->sa_family) {
+	case AF_INET:  return sizeof(struct sockaddr_in);
+	case AF_INET6: return sizeof(struct sockaddr_in6);
+	default:       return kr_error(EINVAL);
+	}
+}
+
 uint16_t kr_inaddr_port(const struct sockaddr *addr)
 {
 	if (!addr) {
@@ -302,6 +361,31 @@ uint16_t kr_inaddr_port(const struct sockaddr *addr)
 	}
 }
 
+int kr_inaddr_str(const struct sockaddr *addr, char *buf, size_t *buflen)
+{
+	int ret = kr_ok();
+	if (!addr || !buf || !buflen) {
+		return kr_error(EINVAL);
+	}
+
+	char str[INET6_ADDRSTRLEN + 6];
+	if (!inet_ntop(addr->sa_family, kr_inaddr(addr), str, sizeof(str))) {
+		return kr_error(errno);
+	}
+	int len = strlen(str);
+	str[len] = '#';
+	u16tostr((uint8_t *)&str[len + 1], kr_inaddr_port(addr));
+	len += 6;
+	str[len] = 0;
+	if (len >= *buflen) {
+		ret = kr_error(ENOSPC);
+	} else {
+		memcpy(buf, str, len + 1);
+	}
+	*buflen = len;
+	return ret;
+}
+
 int kr_straddr_family(const char *addr)
 {
 	if (!addr) {
@@ -361,7 +445,7 @@ int kr_straddr_subnet(void *dst, const char *addr)
 	if (subnet) {
 		*subnet = '\0';
 		subnet += 1;
-		bit_len = atoi(subnet);
+		bit_len = strtol(subnet, NULL, 10);
 		/* Check client subnet length */
 		const int max_len = (family == AF_INET6) ? 128 : 32;
 		if (bit_len < 0 || bit_len > max_len) {
@@ -380,6 +464,84 @@ int kr_straddr_subnet(void *dst, const char *addr)
 	return bit_len;
 }
 
+int kr_straddr_split(const char *addr, char *buf, size_t buflen, uint16_t *port)
+{
+	const int base = 10;
+	long p = 0;
+	size_t addrlen = strlen(addr);
+	char *p_start = strchr(addr, '@');
+	char *p_end;
+
+	if (!p_start) {
+		p_start = strchr(addr, '#');
+	}
+
+	if (p_start) {
+		if (p_start[1] != '\0'){
+			p = strtol(p_start + 1, &p_end, base);
+			if (*p_end != '\0' || p <= 0 || p > UINT16_MAX) {
+				return kr_error(EINVAL);
+			}
+		}
+		addrlen = p_start - addr;
+	}
+
+	/* Check if address is valid. */
+	if (addrlen >= INET6_ADDRSTRLEN) {
+		return kr_error(EINVAL);
+	}
+
+	char str[INET6_ADDRSTRLEN];
+	struct sockaddr_storage ss;
+
+	memcpy(str, addr, addrlen); str[addrlen] = '\0';
+
+	int family = kr_straddr_family(str);
+	if (family == kr_error(EINVAL) || !inet_pton(family, str, &ss)) {
+		return kr_error(EINVAL);
+	}
+
+	/* Address and port contains valid values, return it to caller */
+	if (buf) {
+		if (addrlen >= buflen) {
+			return kr_error(ENOSPC);
+		}
+		memcpy(buf, addr, addrlen); buf[addrlen] = '\0';
+	}
+	if (port) {
+		*port = (uint16_t)p;
+	}
+
+	return kr_ok();
+}
+
+int kr_straddr_join(const char *addr, uint16_t port, char *buf, size_t *buflen)
+{
+	if (!addr || !buf || !buflen) {
+		return kr_error(EINVAL);
+	}
+
+	struct sockaddr_storage ss;
+	int family = kr_straddr_family(addr);
+	if (family == kr_error(EINVAL) || !inet_pton(family, addr, &ss)) {
+		return kr_error(EINVAL);
+	}
+
+	int len = strlen(addr);
+	if (len + 6 >= *buflen) {
+		return kr_error(ENOSPC);
+	}
+
+	memcpy(buf, addr, len + 1);
+	buf[len] = '#';
+	u16tostr((uint8_t *)&buf[len + 1], port);
+	len += 6;
+	buf[len] = 0;
+	*buflen = len;
+
+	return kr_ok();
+}
+
 int kr_bitcmp(const char *a, const char *b, int bits)
 {
 	/* We're using the function from lua directly, so at least for now
@@ -529,13 +691,16 @@ int kr_ranked_rrarray_add(ranked_rr_array_t *array, const knot_rrset_t *rr,
 			continue;
 		}
 		/* Found the entry to merge with.  Check consistency and merge. */
-		bool ok = stashed->rank == rank
-			&& !stashed->cached
-			&& stashed->to_wire == to_wire;
+		bool ok = stashed->rank == rank && !stashed->cached;
 		if (!ok) {
 			assert(false);
 			return kr_error(EEXIST);
 		}
+		/* It may happen that an RRset is first considered useful
+		 * (to_wire = false, e.g. due to being part of glue),
+		 * and later we may find we also want it in the answer. */
+		stashed->to_wire = stashed->to_wire || to_wire;
+
 		return knot_rdataset_merge(&stashed->rr->rrs, &rr->rrs, pool);
 	}
 
@@ -573,13 +738,17 @@ int kr_ranked_rrarray_add(ranked_rr_array_t *array, const knot_rrset_t *rr,
 }
 
 int kr_ranked_rrarray_set_wire(ranked_rr_array_t *array, bool to_wire,
-			       uint32_t qry_uid, bool check_dups)
+			       uint32_t qry_uid, bool check_dups,
+			       bool (*extraCheck)(const ranked_rr_array_entry_t *))
 {
 	for (size_t i = 0; i < array->len; ++i) {
 		ranked_rr_array_entry_t *entry = array->at[i];
 		if (entry->qry_uid != qry_uid) {
 			continue;
 		}
+		if (extraCheck != NULL && !extraCheck(entry)) {
+			continue;
+		}
 		entry->to_wire = to_wire;
 		if (check_dups) {
 			int ret = to_wire_ensure_unique(array, i);
@@ -618,21 +787,6 @@ char *kr_module_call(struct kr_context *ctx, const char *module, const char *pro
 	return NULL;
 }
 
-void kr_rrset_print(const knot_rrset_t *rr, const char *prefix)
-{
-#if KNOT_VERSION_HEX < ((2 << 16) | (4 << 8))
-	char rrtext[KNOT_DNAME_MAXLEN * 2] = {0};
-	knot_rrset_txt_dump(rr, rrtext, sizeof(rrtext), &KNOT_DUMP_STYLE_DEFAULT);
-	kr_log_verbose("%s%s", prefix, rrtext);
-#else
-	size_t size = 4000;
-	char *rrtext = malloc(size);
-	knot_rrset_txt_dump(rr, &rrtext, &size, &KNOT_DUMP_STYLE_DEFAULT);
-	kr_log_verbose("%s%s", prefix, rrtext);
-	free(rrtext);
-#endif
-}
-
 static void flags_to_str(char *dst, const knot_pkt_t *pkt, size_t maxlen)
 {
 	int offset = 0;
@@ -664,7 +818,7 @@ static void flags_to_str(char *dst, const knot_pkt_t *pkt, size_t maxlen)
 	dst[offset] = 0;
 }
 
-static void print_section_opt(const knot_rrset_t *rr, const uint8_t rcode)
+static char *print_section_opt(struct mempool *mp, char *endp, const knot_rrset_t *rr, const uint8_t rcode)
 {
 	uint8_t ercode = knot_edns_get_ext_rcode(rr);
 	uint16_t ext_rcode_id = knot_edns_whole_rcode(ercode, rcode);
@@ -680,18 +834,27 @@ static void print_section_opt(const knot_rrset_t *rr, const uint8_t rcode)
 		}
 	}
 
-	kr_log_verbose(";; EDNS PSEUDOSECTION:\n;; "
-		       "Version: %u; flags: %s; UDP size: %u B; ext-rcode: %s\n\n",
-		       knot_edns_get_version(rr),
-		       (knot_edns_do(rr) != 0) ? "do" : "",
-		       knot_edns_get_payload(rr),
-		       ext_rcode_str);
+	return mp_printf_append(mp, endp,
+		";; EDNS PSEUDOSECTION:\n;; "
+		"Version: %u; flags: %s; UDP size: %u B; ext-rcode: %s\n\n",
+		knot_edns_get_version(rr),
+		(knot_edns_do(rr) != 0) ? "do" : "",
+		knot_edns_get_payload(rr),
+		ext_rcode_str);
 
 }
 
-void kr_pkt_print(knot_pkt_t *pkt)
+char *kr_pkt_text(const knot_pkt_t *pkt)
 {
-	char *snames[] = {";; ANSWER SECTION",";; AUTHORITY SECTION",";; ADDITIONAL SECTION"};
+	if (!pkt) {
+		return NULL;
+	}
+
+	struct mempool *mp = mp_new(512);
+
+	static const char * snames[] = {
+		";; ANSWER SECTION", ";; AUTHORITY SECTION", ";; ADDITIONAL SECTION"
+	};
 	char rrtype[32];
 	char flags[32];
 	char qname[KNOT_DNAME_MAXLEN];
@@ -702,6 +865,7 @@ void kr_pkt_print(knot_pkt_t *pkt)
 	const knot_lookup_t *rcode = knot_lookup_by_id(knot_rcode_names, pkt_rcode);
 	const knot_lookup_t *opcode = knot_lookup_by_id(knot_opcode_names, pkt_opcode);
 	uint16_t qry_id = knot_wire_get_id(pkt->wire);
+	uint16_t qdcount = knot_wire_get_qdcount(pkt->wire);
 
 	if (rcode != NULL) {
 		rcode_str = rcode->name;
@@ -710,73 +874,114 @@ void kr_pkt_print(knot_pkt_t *pkt)
 		opcode_str = opcode->name;
 	}
 	flags_to_str(flags, pkt, sizeof(flags));
-	knot_dname_to_str(qname, knot_pkt_qname(pkt), KNOT_DNAME_MAXLEN);
-	knot_rrtype_to_string(knot_pkt_qtype(pkt), rrtype, sizeof(rrtype));
-	kr_log_verbose(";; ->>HEADER<<- opcode: %s; status: %s; id: %hu\n",
-		       opcode_str, rcode_str, qry_id);
 
-	kr_log_verbose(";; Flags: %s QUERY: %hu; ANSWER: %hu; "
-		       "AUTHORITY: %hu; ADDITIONAL: %hu\n\n",
-		       flags,
-		       knot_wire_get_qdcount(pkt->wire),
-		       knot_wire_get_ancount(pkt->wire),
-		       knot_wire_get_nscount(pkt->wire),
-		       knot_wire_get_arcount(pkt->wire));
+	char *ptr = mp_printf(mp,
+		";; ->>HEADER<<- opcode: %s; status: %s; id: %hu\n"
+		";; Flags: %s QUERY: %hu; ANSWER: %hu; "
+		"AUTHORITY: %hu; ADDITIONAL: %hu\n\n",
+		opcode_str, rcode_str, qry_id,
+		flags,
+		qdcount,
+		knot_wire_get_ancount(pkt->wire),
+		knot_wire_get_nscount(pkt->wire),
+		knot_wire_get_arcount(pkt->wire));
 
 	if (knot_pkt_has_edns(pkt)) {
-		print_section_opt(pkt->opt_rr,
-		                  knot_wire_get_rcode(pkt->wire));
+		ptr = print_section_opt(mp, ptr, pkt->opt_rr, knot_wire_get_rcode(pkt->wire));
 	}
 
-	kr_log_verbose(";; QUESTION SECTION\n%s\t\t%s\n\n", qname, rrtype);
-	for (knot_section_t i = KNOT_ANSWER; i <= KNOT_AUTHORITY; ++i) {
+	if (qdcount == 1) {
+		knot_dname_to_str(qname, knot_pkt_qname(pkt), KNOT_DNAME_MAXLEN);
+		knot_rrtype_to_string(knot_pkt_qtype(pkt), rrtype, sizeof(rrtype));
+		ptr = mp_printf_append(mp, ptr, ";; QUESTION SECTION\n%s\t\t%s\n", qname, rrtype);
+	} else if (qdcount > 1) {
+		ptr = mp_printf_append(mp, ptr, ";; Warning: unsupported QDCOUNT %hu\n", qdcount);
+	}
+
+	for (knot_section_t i = KNOT_ANSWER; i <= KNOT_ADDITIONAL; ++i) {
 		const knot_pktsection_t *sec = knot_pkt_section(pkt, i);
-		if (sec->count == 0) {
+		if (sec->count == 0 || knot_pkt_rr(sec, 0)->type == KNOT_RRTYPE_OPT) {
+			/* OPT RRs are _supposed_ to be the last ^^, if they appear */
 			continue;
 		}
-		kr_log_verbose("%s\n", snames[i - KNOT_ANSWER]);
+
+		ptr = mp_printf_append(mp, ptr, "\n%s\n", snames[i - KNOT_ANSWER]);
 		for (unsigned k = 0; k < sec->count; ++k) {
 			const knot_rrset_t *rr = knot_pkt_rr(sec, k);
-			kr_rrset_print(rr, "");
+			if (rr->type == KNOT_RRTYPE_OPT) {
+				continue;
+			}
+			auto_free char *rr_text = kr_rrset_text(rr);
+			ptr = mp_printf_append(mp, ptr, "%s", rr_text);
 		}
-		kr_log_verbose("\n");
 	}
-	const knot_pktsection_t *sec = knot_pkt_section(pkt, KNOT_ADDITIONAL);
-	bool header_was_printed = false;
-	for (unsigned k = 0; k < sec->count; ++k) {
-		const knot_rrset_t *rr = knot_pkt_rr(sec, k);
-		if (rr->type == KNOT_RRTYPE_OPT) {
-			continue;
-		}
-		if (!header_was_printed) {
-			header_was_printed = true;
-			kr_log_verbose("%s\n", snames[KNOT_ADDITIONAL - KNOT_ANSWER]);
-		}
-		kr_rrset_print(rr, "");
+
+	/* Close growing buffer and duplicate result before deleting */
+	char *result = strdup(ptr);
+	mp_delete(mp);
+	return result;
+}
+
+char *kr_rrset_text(const knot_rrset_t *rr)
+{
+	if (!rr) {
+		return NULL;
 	}
-	kr_log_verbose("\n");
+
+	/* Note: knot_rrset_txt_dump will double the size until the rrset fits */
+	size_t bufsize = 128;
+	char *buf = malloc(bufsize);
+	int ret = knot_rrset_txt_dump(rr, &buf, &bufsize, &KNOT_DUMP_STYLE_DEFAULT);
+	if (ret < 0) {
+		free(buf);
+		return NULL;
+	}
+
+	return buf;
 }
 
-void kr_dname_print(const knot_dname_t *name, const char *prefix, const char *postfix)
+uint64_t kr_now()
 {
-	char str[KNOT_DNAME_MAXLEN] = {0};
-	knot_dname_to_str(str, name, KNOT_DNAME_MAXLEN);
-	kr_log_verbose("%s%s%s", prefix, str, postfix);
+	return uv_now(uv_default_loop());
 }
 
-void kr_rrtype_print(const uint16_t rrtype, const char *prefix, const char *postfix)
+int knot_dname_lf2wire(knot_dname_t * const dst, uint8_t len, const uint8_t *lf)
 {
-	char str[32] = {0};
-	knot_rrtype_to_string(rrtype, str, 32);
-	kr_log_verbose("%s%s%s", prefix, str, postfix);
+	knot_dname_t *d = dst; /* moving "cursor" as we write it out */
+	bool ok = d && (len == 0 || lf);
+	if (!ok) {
+		assert(false);
+		return kr_error(EINVAL);
+	}
+	/* we allow the final zero byte to be omitted */
+	if (!len) {
+		goto finish;
+	}
+	if (lf[len - 1]) {
+		++len;
+	}
+	/* convert the name, one label at a time */
+	int label_end = len - 1; /* index of the zero byte after the current label */
+	while (label_end >= 0) {
+		/* find label_start */
+		int i = label_end - 1;
+		while (i >= 0 && lf[i])
+			--i;
+		const int label_start = i + 1; /* index of the first byte of the current label */
+		const int label_len = label_end - label_start;
+		assert(label_len >= 0);
+		if (label_len > 63 || label_len <= 0)
+			return kr_error(EILSEQ);
+		/* write the label */
+		*d = label_len;
+		++d;
+		memcpy(d, lf + label_start, label_len);
+		d += label_len;
+		/* next label */
+		label_end = label_start - 1;
+	}
+finish:
+	*d = 0; /* the final zero */
+	++d;
+	return d - dst;
 }
-
-void kr_qry_print(const struct kr_query *qry, const char *prefix, const char *postfix)
-{
-	char str[6] = {0};
-	knot_rrclass_to_string(qry->sclass, str, sizeof(str));
-	kr_dname_print(qry->sname, prefix, " ");
-	kr_log_verbose("%s",str);
-	kr_rrtype_print(qry->stype, " ", postfix);
-}
-
