@@ -26,15 +26,23 @@
 #include <libknot/packet/pkt.h>
 #include <libknot/rrset.h>
 #include <libknot/rrtype/rrsig.h>
+#include <lua.h>
 #include "lib/generic/map.h"
 #include "lib/generic/array.h"
 #include "lib/defines.h"
 
 struct kr_query;
+struct kr_request;
 
 /*
  * Logging and debugging.
  */
+
+/** @brief Callback for request events. */
+typedef void (*trace_callback_f)(struct kr_request *request);
+/** @brief Callback for request logging handler. */
+typedef void (*trace_log_f)(const struct kr_query *query, const char *source, const char *msg);
+
 #define kr_log_info(fmt, ...) do { printf((fmt), ## __VA_ARGS__); fflush(stdout); } while(0)
 #define kr_log_error(fmt, ...) fprintf(stderr, (fmt), ## __VA_ARGS__)
 
@@ -46,7 +54,23 @@ KR_EXPORT extern bool kr_verbose_status;
 KR_EXPORT bool kr_verbose_set(bool status);
 
 /** Log a message if in --verbose mode. */
-KR_EXPORT void kr_log_verbose(const char *fmt, ...);
+KR_EXPORT KR_PRINTF(1)
+void kr_log_verbose(const char *fmt, ...);
+
+/**
+ * @brief Return true if the query has request log handler installed.
+ */
+#define kr_log_trace_enabled(query) ((query) && (query)->request && (query)->request->trace_log)
+
+/**
+ * Log a message through the request log handler.
+ * @param  query current query
+ * @param  source message source
+ * @param  fmt message format
+ * @return true if the message was logged
+ */
+KR_EXPORT KR_PRINTF(3)
+bool kr_log_trace(const struct kr_query *query, const char *source, const char *fmt, ...);
 
 #ifdef NOVERBOSELOG
 /* Efficient compile-time disabling of verbose messages. */
@@ -55,8 +79,9 @@ KR_EXPORT void kr_log_verbose(const char *fmt, ...);
 #endif
 
 /** Block run in --verbose mode; optimized when not run. */
-#define WITH_VERBOSE if(__builtin_expect(kr_verbose_status, false))
-#define kr_log_verbose WITH_VERBOSE kr_log_verbose
+#define VERBOSE_STATUS __builtin_expect(kr_verbose_status, false)
+#define WITH_VERBOSE(query) if(__builtin_expect(kr_verbose_status || kr_log_trace_enabled(query), false))
+#define kr_log_verbose if(VERBOSE_STATUS) kr_log_verbose
 
 
 /* C11 compatibility, but without any implementation so far. */
@@ -120,9 +145,10 @@ struct ranked_rr_array_entry {
 	uint32_t qry_uid;
 	uint8_t rank; /**< enum kr_rank */
 	uint8_t revalidation_cnt;
-	bool cached;  /**< whether it has been stashed to cache already */
-	bool yielded;
-	bool to_wire; /**< whether to be put into the answer */
+	bool cached : 1;  /**< whether it has been stashed to cache already */
+	bool yielded : 1;
+	bool to_wire : 1; /**< whether to be put into the answer */
+	bool expiring : 1; /**< low remaining TTL; see is_expiring; only used in cache ATM */
 	knot_rrset_t *rr;
 };
 typedef struct ranked_rr_array_entry ranked_rr_array_entry_t;
@@ -190,16 +216,22 @@ const char *kr_inaddr(const struct sockaddr *addr);
 /** Address family. */
 KR_EXPORT KR_PURE
 int kr_inaddr_family(const struct sockaddr *addr);
-/** Address length for given family. */
+/** Address length for given family, i.e. sizeof(struct in*_addr). */
 KR_EXPORT KR_PURE
 int kr_inaddr_len(const struct sockaddr *addr);
+/** Sockaddr length for given family, i.e. sizeof(struct sockaddr_in*). */
+KR_EXPORT KR_PURE
+int kr_sockaddr_len(const struct sockaddr *addr);
 /** Port. */
 KR_EXPORT KR_PURE
 uint16_t kr_inaddr_port(const struct sockaddr *addr);
+/** String representation for given address as "<addr>#<port>" */
+KR_EXPORT
+int kr_inaddr_str(const struct sockaddr *addr, char *buf, size_t *buflen);
 /** Return address type for string. */
 KR_EXPORT KR_PURE
 int kr_straddr_family(const char *addr);
-/** Return address length in given family. */
+/** Return address length in given family (struct in*_addr). */
 KR_EXPORT KR_CONST
 int kr_family_len(int family);
 /** Create a sockaddr* from string+port representation (also accepts IPv6 link-local). */
@@ -210,6 +242,26 @@ struct sockaddr * kr_straddr_socket(const char *addr, int port);
 KR_EXPORT
 int kr_straddr_subnet(void *dst, const char *addr);
 
+/** Splits ip address specified as "addr@port" or "addr#port" into addr and port
+  * and performs validation.
+  * @note if #port part isn't present, then port will be set to 0.
+  *       buf and\or port can be set to NULL.
+  * @return kr_error(EINVAL) - addr part doesn't contains valid ip address or
+  *                            #port part is out-of-range (either < 0 either > UINT16_MAX)
+  *         kr_error(ENOSP)  - buflen is too small
+  */
+KR_EXPORT
+int kr_straddr_split(const char *addr, char *buf, size_t buflen, uint16_t *port);
+/** Formats ip address and port in "addr#port" format.
+  * and performs validation.
+  * @note Port always formatted as five-character string with leading zeros.
+  * @return kr_error(EINVAL) - addr or buf is NULL or buflen is 0 or
+  *                            addr doesn't contain a valid ip address
+  *         kr_error(ENOSP)  - buflen is too small
+  */
+KR_EXPORT
+int kr_straddr_join(const char *addr, uint16_t port, char *buf, size_t *buflen);
+
 /** Compare memory bitwise.  The semantics is "the same" as for memcmp().
  *  The partial byte is considered with more-significant bits first,
  *  so this is e.g. suitable for comparing IP prefixes. */
@@ -252,16 +304,30 @@ int kr_ranked_rrarray_add(ranked_rr_array_t *array, const knot_rrset_t *rr,
  * @param to_wire Records must be\must not be recorded in final answer.
  * @param qry_uid Query uid.
  * @param check_dups When to_wire is true, try to avoid duplicate RRSets.
+ * @param extraCheck optional function checking whether to consider the record
  * @return 0 or an error
  */
 int kr_ranked_rrarray_set_wire(ranked_rr_array_t *array, bool to_wire,
-			       uint32_t qry_uid, bool check_dups);
+			       uint32_t qry_uid, bool check_dups,
+			       bool (*extraCheck)(const ranked_rr_array_entry_t *));
 
-void kr_rrset_print(const knot_rrset_t *rr, const char *prefix);
-void kr_qry_print(const struct kr_query *qry, const char *prefix, const char *postfix);
-void kr_pkt_print(knot_pkt_t *pkt);
-void kr_dname_print(const knot_dname_t *name, const char *prefix, const char *postfix);
-void kr_rrtype_print(const uint16_t rrtype, const char *prefix, const char *postfix);
+KR_PURE
+char *kr_pkt_text(const knot_pkt_t *pkt);
+
+KR_PURE
+char *kr_rrset_text(const knot_rrset_t *rr);
+
+KR_PURE
+static inline char *kr_dname_text(const knot_dname_t *name) {
+	return knot_dname_to_str_alloc(name);
+}
+
+KR_CONST
+static inline char *kr_rrtype_text(const uint16_t rrtype) {
+	char type_str[32] = {0};
+	knot_rrtype_to_string(rrtype, type_str, sizeof(type_str));
+	return strdup(type_str);
+}
 
 /**
  * Call module property.
@@ -286,3 +352,66 @@ static inline uint16_t kr_rrset_type_maysig(const knot_rrset_t *rr)
 		type = knot_rrsig_type_covered(&rr->rrs, 0);
 	return type;
 }
+
+/** Printf onto the lua stack, avoiding additional copy (thin wrapper). */
+KR_PRINTF(2)
+static inline const char *lua_push_printf(lua_State *L, const char *fmt, ...)
+{
+	va_list args;
+	va_start(args, fmt);
+	const char *ret = lua_pushvfstring(L, fmt, args);
+	va_end(args);
+	return ret;
+}
+
+/** @internal Return string representation of addr.
+ *  @note return pointer to static string
+ */
+static inline char *kr_straddr(const struct sockaddr *addr)
+{
+	assert(addr != NULL);
+	/* We are the sinle-threaded application */
+	static char str[INET6_ADDRSTRLEN + 6];
+	size_t len = sizeof(str);
+	int ret = kr_inaddr_str(addr, str, &len);
+	return ret != kr_ok() || len == 0 ? NULL : str;
+}
+
+/** The current time in monotonic milliseconds.
+ *
+ * \note it may be outdated in case of long callbacks; see uv_now().
+ */
+KR_EXPORT
+uint64_t kr_now();
+
+/** Convert name from lookup format to wire.  See knot_dname_lf
+ *
+ * \note len bytes are read and len+1 are written with *normal* LF,
+ * 	 but it's also allowed that the final zero byte is omitted in LF.
+ * \return the number of bytes written (>0) or error code (<0)
+ */
+int knot_dname_lf2wire(knot_dname_t *dst, uint8_t len, const uint8_t *lf);
+
+/** Patched knot_dname_lf.  LF for "." has length zero instead of one, for consistency.
+ * (TODO: consistency?)
+ * \param add_wildcard append the wildcard label
+ * \note packet is always NULL
+ */
+static inline int kr_dname_lf(uint8_t *dst, const knot_dname_t *src, bool add_wildcard)
+{
+	int ret = knot_dname_lf(dst, src, NULL);
+	if (ret)
+		return ret;
+	int len = dst[0];
+	if (len == 1)
+		len = 0;
+	if (add_wildcard) {
+		if (len + 2 > KNOT_DNAME_MAXLEN)
+			return kr_error(ENOSPC);
+		dst[len + 1] = '*';
+		dst[len + 2] = '\0';
+		len += 2;
+	}
+	dst[0] = len;
+	return KNOT_EOK;
+};
