@@ -22,11 +22,14 @@
 
 #include "lib/zonecut.h"
 #include "lib/rplan.h"
+#include "contrib/cleanup.h"
 #include "lib/defines.h"
 #include "lib/layer.h"
 #include "lib/resolve.h"
 #include "lib/generic/pack.h"
 
+#define VERBOSE_MSG(qry, fmt...) QRVERBOSE(qry, "zcut", fmt)
+
 /* Root hint descriptor. */
 struct hint_info {
 	const knot_dname_t *name;
@@ -36,49 +39,6 @@ struct hint_info {
 
 #define U8(x) (const uint8_t *)(x)
 
-/* Initialize with SBELT name servers. */
-#include "lib/root-hints.inc"
-#define HINT_LEN_A sizeof(struct in_addr)
-#define HINT_LEN_AAAA sizeof(struct in6_addr)
-#define SBELT_LEN (sizeof(SBELT) / sizeof(SBELT[0]))
-static const struct hint_info SBELT[] = {
-	#define HINT(name,type) { \
-		U8("\x01" #name "\x0c""root-servers""\x03""net"), \
-		HINT_LEN_##type, \
-		U8(HINT_##name##_##type), \
-      	}
-	HINT(j, A),
-	HINT(k, A),
-	HINT(d, A),
-	HINT(e, A),
-	HINT(f, A),
-	HINT(g, A),
-	HINT(h, A),
-	HINT(i, A),
-	HINT(l, A),
-	HINT(m, A),
-	HINT(b, A),
-	HINT(c, A),
-	HINT(a, A),
-	HINT(a, AAAA),
-	HINT(b, AAAA),
-	HINT(c, AAAA),
-	HINT(d, AAAA),
-	HINT(e, AAAA),
-	HINT(f, AAAA),
-	HINT(g, AAAA),
-	HINT(h, AAAA),
-	HINT(i, AAAA),
-	HINT(j, AAAA),
-	HINT(k, AAAA),
-	HINT(l, AAAA),
-	HINT(m, AAAA),
-	#undef HINT
-};
-#undef HINT_LEN_A
-#undef HINT_LEN_AAAA
-
-
 static void update_cut_name(struct kr_zonecut *cut, const knot_dname_t *name)
 {
 	if (knot_dname_is_equal(name, cut->name)) {
@@ -298,8 +258,6 @@ int kr_zonecut_set_sbelt(struct kr_context *ctx, struct kr_zonecut *cut)
 	if (!ctx || !cut) {
 		return kr_error(EINVAL);
 	}
-	/* @warning _NOT_ thread-safe */
-	static knot_rdata_t rdata_arr[RDATA_ARR_MAX];
 
 	update_cut_name(cut, U8(""));
 	map_walk(&cut->nsset, free_addr_set, cut->pool);
@@ -309,48 +267,44 @@ int kr_zonecut_set_sbelt(struct kr_context *ctx, struct kr_zonecut *cut)
 	int ret = 0;
 	if (ctx->root_hints.nsset.root) {
 		ret = kr_zonecut_copy(cut, &ctx->root_hints);
-	} else {
-		/* Copy compiled-in root hints */
-		for (unsigned i = 0; i < SBELT_LEN; ++i) {
-			const struct hint_info *hint = &SBELT[i];
-			knot_rdata_init(rdata_arr, hint->len, hint->addr, 0);
-			ret = kr_zonecut_add(cut, hint->name, rdata_arr);
-			if (ret != 0) {
-				break;
-			}
-		}
 	}
 	return ret;
 }
 
-/** Fetch address for zone cut. */
-static void fetch_addr(struct kr_zonecut *cut, struct kr_cache *cache, const knot_dname_t *ns, uint16_t rrtype, uint32_t timestamp)
+/** Fetch address for zone cut.  Any rank is accepted (i.e. glue as well). */
+static void fetch_addr(struct kr_zonecut *cut, struct kr_cache *cache,
+			const knot_dname_t *ns, uint16_t rrtype,
+			const struct kr_query *qry)
+// LATER(optim.): excessive data copying
 {
-	uint8_t rank = 0;
-	knot_rrset_t cached_rr;
-	knot_rrset_init(&cached_rr, (knot_dname_t *)ns, rrtype, KNOT_CLASS_IN);
-	if (kr_cache_peek_rr(cache, &cached_rr, &rank, NULL, &timestamp) != 0) {
+	struct kr_cache_p peek;
+	if (kr_cache_peek_exact(cache, ns, rrtype, &peek) != 0) {
+		return;
+	}
+	int32_t new_ttl = kr_cache_ttl(&peek, qry, ns, rrtype);
+	if (new_ttl < 0) {
 		return;
 	}
 
+	knot_rrset_t cached_rr;
+	knot_rrset_init(&cached_rr, /*const-cast*/(knot_dname_t *)ns, rrtype, KNOT_CLASS_IN);
+	if (kr_cache_materialize(&cached_rr.rrs, &peek, new_ttl, cut->pool) < 0) {
+		return;
+	}
 	knot_rdata_t *rd = cached_rr.rrs.data;
 	for (uint16_t i = 0; i < cached_rr.rrs.rr_count; ++i) {
-		if (knot_rdata_ttl(rd) > timestamp) {
-			(void) kr_zonecut_add(cut, ns, rd);
-		}
+		(void) kr_zonecut_add(cut, ns, rd);
 		rd = kr_rdataset_next(rd);
 	}
 }
 
 /** Fetch best NS for zone cut. */
 static int fetch_ns(struct kr_context *ctx, struct kr_zonecut *cut,
-		    const knot_dname_t *name, uint32_t timestamp,
-		    uint8_t * restrict rank, uint8_t * restrict flags)
+		    const knot_dname_t *name, const struct kr_query *qry,
+		    uint8_t * restrict rank)
 {
-	uint32_t drift = timestamp;
-	knot_rrset_t cached_rr;
-	knot_rrset_init(&cached_rr, (knot_dname_t *)name, KNOT_RRTYPE_NS, KNOT_CLASS_IN);
-	int ret = kr_cache_peek_rr(&ctx->cache, &cached_rr, rank, flags, &drift);
+	struct kr_cache_p peek;
+	int ret = kr_cache_peek_exact(&ctx->cache, name, KNOT_RRTYPE_NS, &peek);
 	if (ret != 0) {
 		return ret;
 	}
@@ -358,32 +312,37 @@ static int fetch_ns(struct kr_context *ctx, struct kr_zonecut *cut,
 	 * completely untrustworthy could get into the cache, e.g out-of-bailiwick
 	 * records that weren't validated.
 	 */
+	*rank = peek.rank;
 
-	/* Materialize as we'll going to do more cache lookups. */
-	knot_rrset_t rr_copy;
-	ret = kr_cache_materialize(&rr_copy, &cached_rr, drift, 0, cut->pool);
-	if (ret != 0) {
+	int32_t new_ttl = kr_cache_ttl(&peek, qry, name, KNOT_RRTYPE_NS);
+	if (new_ttl < 0) {
+		return kr_error(ESTALE);
+	}
+	/* Materialize the rdataset temporarily, for simplicity. */
+	knot_rdataset_t ns_rds = { 0, NULL };
+	ret = kr_cache_materialize(&ns_rds, &peek, new_ttl, cut->pool);
+	if (ret < 0) {
 		return ret;
 	}
 
 	/* Insert name servers for this zone cut, addresses will be looked up
 	 * on-demand (either from cache or iteratively) */
-	for (unsigned i = 0; i < rr_copy.rrs.rr_count; ++i) {
-		const knot_dname_t *ns_name = knot_ns_name(&rr_copy.rrs, i);
-		kr_zonecut_add(cut, ns_name, NULL);
+	for (unsigned i = 0; i < ns_rds.rr_count; ++i) {
+		const knot_dname_t *ns_name = knot_ns_name(&ns_rds, i);
+		(void) kr_zonecut_add(cut, ns_name, NULL);
 		/* Fetch NS reputation and decide whether to prefetch A/AAAA records. */
 		unsigned *cached = lru_get_try(ctx->cache_rep,
 				(const char *)ns_name, knot_dname_size(ns_name));
 		unsigned reputation = (cached) ? *cached : 0;
-		if (!(reputation & KR_NS_NOIP4) && !(ctx->options & QUERY_NO_IPV4)) {
-			fetch_addr(cut, &ctx->cache, ns_name, KNOT_RRTYPE_A, timestamp);
+		if (!(reputation & KR_NS_NOIP4) && !(ctx->options.NO_IPV4)) {
+			fetch_addr(cut, &ctx->cache, ns_name, KNOT_RRTYPE_A, qry);
 		}
-		if (!(reputation & KR_NS_NOIP6) && !(ctx->options & QUERY_NO_IPV6)) {
-			fetch_addr(cut,  &ctx->cache, ns_name, KNOT_RRTYPE_AAAA, timestamp);
+		if (!(reputation & KR_NS_NOIP6) && !(ctx->options.NO_IPV6)) {
+			fetch_addr(cut,  &ctx->cache, ns_name, KNOT_RRTYPE_AAAA, qry);
 		}
 	}
 
-	knot_rrset_clear(&rr_copy, cut->pool);
+	knot_rdataset_clear(&ns_rds, cut->pool);
 	return kr_ok();
 }
 
@@ -391,33 +350,41 @@ static int fetch_ns(struct kr_context *ctx, struct kr_zonecut *cut,
  * Fetch secure RRSet of given type.
  */
 static int fetch_secure_rrset(knot_rrset_t **rr, struct kr_cache *cache,
-	const knot_dname_t *owner, uint16_t type, knot_mm_t *pool, uint32_t timestamp)
+	const knot_dname_t *owner, uint16_t type, knot_mm_t *pool,
+	const struct kr_query *qry)
 {
 	if (!rr) {
-		return kr_error(ENOENT);
+		assert(!EINVAL);
+		return kr_error(EINVAL);
 	}
-
-	uint8_t rank = 0;
-	uint32_t drift = timestamp;
-	knot_rrset_t cached_rr;
-	knot_rrset_init(&cached_rr, (knot_dname_t *)owner, type, KNOT_CLASS_IN);
-	int ret = kr_cache_peek_rr(cache, &cached_rr, &rank, NULL, &drift);
+	/* peek, check rank and TTL */
+	struct kr_cache_p peek;
+	int ret = kr_cache_peek_exact(cache, owner, type, &peek);
 	if (ret != 0) {
 		return ret;
 	}
-	const bool rankOK = kr_rank_test(rank, KR_RANK_SECURE);
-	if (!rankOK) {
+	if (!kr_rank_test(peek.rank, KR_RANK_SECURE)) {
 		return kr_error(ENOENT);
 	}
-
+	int32_t new_ttl = kr_cache_ttl(&peek, qry, owner, type);
+	if (new_ttl < 0) {
+		return kr_error(ESTALE);
+	}
+	/* materialize a new RRset */
 	knot_rrset_free(rr, pool);
 	*rr = mm_alloc(pool, sizeof(knot_rrset_t));
 	if (*rr == NULL) {
 		return kr_error(ENOMEM);
 	}
-
-	ret = kr_cache_materialize(*rr, &cached_rr, drift, 0, pool);
-	if (ret != 0) {
+	owner = knot_dname_copy(/*const-cast*/(knot_dname_t *)owner, pool);
+	if (!owner) {
+		mm_free(pool, *rr);
+		*rr = NULL;
+		return kr_error(ENOMEM);
+	}
+	knot_rrset_init(*rr, /*const-cast*/(knot_dname_t *)owner, type, KNOT_CLASS_IN);
+	ret = kr_cache_materialize(&(*rr)->rrs, &peek, new_ttl, pool);
+	if (ret < 0) {
 		knot_rrset_free(rr, pool);
 		return ret;
 	}
@@ -425,9 +392,11 @@ static int fetch_secure_rrset(knot_rrset_t **rr, struct kr_cache *cache,
 	return kr_ok();
 }
 
-int kr_zonecut_find_cached(struct kr_context *ctx, struct kr_zonecut *cut, const knot_dname_t *name,
-                           uint32_t timestamp, bool * restrict secured)
+int kr_zonecut_find_cached(struct kr_context *ctx, struct kr_zonecut *cut,
+			   const knot_dname_t *name, const struct kr_query *qry,
+			   bool * restrict secured)
 {
+	//VERBOSE_MSG(qry, "_find_cached\n");
 	if (!ctx || !cut || !name) {
 		return kr_error(EINVAL);
 	}
@@ -441,23 +410,29 @@ int kr_zonecut_find_cached(struct kr_context *ctx, struct kr_zonecut *cut, const
 	while (true) {
 		/* Fetch NS first and see if it's insecure. */
 		uint8_t rank = 0;
-		uint8_t flags = 0;
 		const bool is_root = (label[0] == '\0');
-		if (fetch_ns(ctx, cut, label, timestamp, &rank, &flags) == 0) {
+		if (fetch_ns(ctx, cut, label, qry, &rank) == 0) {
 			/* Flag as insecure if cached as this */
-			if (kr_rank_test(rank, KR_RANK_INSECURE) ||
-			    (flags & KR_CACHE_FLAG_NODS)) {
+			if (kr_rank_test(rank, KR_RANK_INSECURE)) {
 				*secured = false;
 			}
 			/* Fetch DS and DNSKEY if caller wants secure zone cut */
+			int ret_ds = 1, ret_dnskey = 1;
 			if (*secured || is_root) {
-				fetch_secure_rrset(&cut->trust_anchor, &ctx->cache, label,
-					    KNOT_RRTYPE_DS, cut->pool, timestamp);
-				fetch_secure_rrset(&cut->key, &ctx->cache, label,
-					    KNOT_RRTYPE_DNSKEY, cut->pool, timestamp);
+				ret_ds = fetch_secure_rrset(&cut->trust_anchor, &ctx->cache,
+						label, KNOT_RRTYPE_DS, cut->pool, qry);
+				ret_dnskey = fetch_secure_rrset(&cut->key, &ctx->cache,
+						label, KNOT_RRTYPE_DNSKEY, cut->pool, qry);
 			}
 			update_cut_name(cut, label);
 			mm_free(cut->pool, qname);
+			kr_cache_sync(&ctx->cache);
+			WITH_VERBOSE(qry) {
+				auto_free char *label_str = kr_dname_text(label);
+				VERBOSE_MSG(qry,
+					"found cut: %s (return codes: DS %d, DNSKEY %d)\n",
+					label_str, ret_ds, ret_dnskey);
+			}
 			return kr_ok();
 		}
 		/* Subtract label from QNAME. */
@@ -467,6 +442,7 @@ int kr_zonecut_find_cached(struct kr_context *ctx, struct kr_zonecut *cut, const
 			break;
 		}
 	}
+	kr_cache_sync(&ctx->cache);
 	mm_free(cut->pool, qname);
 	return kr_error(ENOENT);
 }
