@@ -19,7 +19,7 @@
 #include "lib/generic/map.h"
 #include "lib/generic/pack.h"
 #include "lib/defines.h"
-#include "lib/cache.h"
+#include "lib/cache/api.h"
 
 struct kr_rplan;
 struct kr_context;
@@ -142,10 +142,11 @@ int kr_zonecut_set_sbelt(struct kr_context *ctx, struct kr_zonecut *cut);
  * @param ctx       resolution context (to fetch data from LRU caches)
  * @param cut       zone cut to be populated
  * @param name      QNAME to start finding zone cut for
- * @param timestamp transaction timestamp
+ * @param qry       query for timestamp and stale-serving decisions
  * @param secured   set to true if want secured zone cut, will return false if it is provably insecure
  * @return 0 or error code (ENOENT if it doesn't find anything)
  */
 KR_EXPORT
-int kr_zonecut_find_cached(struct kr_context *ctx, struct kr_zonecut *cut, const knot_dname_t *name,
-                           uint32_t timestamp, bool * restrict secured);
+int kr_zonecut_find_cached(struct kr_context *ctx, struct kr_zonecut *cut,
+			   const knot_dname_t *name, const struct kr_query *qry,
+			   bool * restrict secured);
