@@ -1,4 +1,4 @@
-/*  Copyright (C) 2016 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2016-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -73,12 +73,12 @@ int cookies_deinit(struct kr_module *module)
 }
 
 KR_EXPORT
-const knot_layer_api_t *cookies_layer(struct kr_module *module)
+const kr_layer_api_t *cookies_layer(struct kr_module *module)
 {
 	/* The function answer_finalize() in resolver is called before any
 	 * .finish callback. Therefore this layer does not use it. */
 
-	static knot_layer_api_t _layer = {
+	static kr_layer_api_t _layer = {
 		.begin = &check_request,
 		.consume = &check_response
 	};
