@@ -27,6 +27,7 @@
 #include <ccan/json/json.h>
 #include <ucw/mempool.h>
 #include <contrib/cleanup.h>
+#include <lauxlib.h>
 
 #include "daemon/engine.h"
 #include "lib/zonecut.h"
@@ -152,8 +153,9 @@ static int query(kr_layer_t *ctx, knot_pkt_t *pkt)
 	}
 
 	VERBOSE_MSG(qry, "<= answered from hints\n");
-	qry->flags &= ~QUERY_DNSSEC_WANT; /* Never authenticated */
-	qry->flags |= QUERY_CACHED|QUERY_NO_MINIMIZE;
+	qry->flags.DNSSEC_WANT = false; /* Never authenticated */
+	qry->flags.CACHED = true;
+	qry->flags.NO_MINIMIZE = true;
 	pkt->parsed = pkt->size;
 	knot_wire_set_qr(pkt->wire);
 	return KR_STATE_DONE;
@@ -564,6 +566,17 @@ static char* hint_root(void *env, struct kr_module *module, const char *args)
 	return pack_hints(root_hints);
 }
 
+static char* hint_root_file(void *env, struct kr_module *module, const char *args)
+{
+	struct engine *engine = env;
+	struct kr_context *ctx = &engine->resolver;
+	const char *err_msg = engine_hint_root_file(ctx, args);
+	if (err_msg) {
+		luaL_error(engine->L, "error when opening '%s': %s", args, err_msg);
+	}
+	return strdup(err_msg ? err_msg : "");
+}
+
 /*
  * Module implementation.
  */
@@ -649,6 +662,7 @@ struct kr_prop *hints_props(void)
 	    { &hint_get,    "get", "Retrieve hint for given name.", },
 	    { &hint_add_hosts, "add_hosts", "Load a file with hosts-like formatting and add contents into hints.", },
 	    { &hint_root,   "root", "Replace root hints set (empty value to return current list).", },
+	    { &hint_root_file, "root_file", "Replace root hints set from a zonefile.", },
 	    { NULL, NULL, NULL }
 	};
 	return prop_list;
