@@ -1,5 +1,4 @@
 local kres = require('kres')
-local bit = require('bit')
 local ffi = require('ffi')
 
 local todname = kres.str2dname -- not available during module load otherwise
@@ -34,25 +33,40 @@ if has_socket then
 	end
 end
 
-local function addr_split_port(target)
+local function addr_split_port(target, default_port)
+	assert(default_port)
+	assert(type(default_port) == 'number')
 	local addr, port = target:match '([^@]*)@?(.*)'
-	port = port and tonumber(port) or 53
-	return addr, port
+	local nport
+	if port ~= "" then
+		nport = tonumber(port)
+		if not nport or nport < 1 or nport > 65535 then
+			error('port "'.. port  ..'" is not valid')
+		end
+	end
+	return addr, nport or default_port
 end
 
 -- String address@port -> sockaddr.
-local function addr2sock(target)
-	local addr, port = addr_split_port(target)
-	sock = ffi.gc(ffi.C.kr_straddr_socket(addr, port), ffi.C.free);
+local function addr2sock(target, default_port)
+	local addr, port = addr_split_port(target, default_port)
+	local sock = ffi.gc(ffi.C.kr_straddr_socket(addr, port), ffi.C.free);
 	if sock == nil then
 		error("target '"..target..'" is not a valid IP address')
 	end
 	return sock
 end
 
+-- policy functions are defined below
+local policy = {}
+
+function policy.PASS(state, _)
+	return state
+end
+
 -- Mirror request elsewhere, and continue solving
-local function mirror(target)
-	local addr, port = addr_split_port(target)
+function policy.MIRROR(target)
+	local addr, port = addr_split_port(target, 53)
 	local sink, err = socket_client(addr, port)
 	if not sink then panic('MIRROR target %s is not a valid: %s', target, err) end
 	return function(state, req)
@@ -65,50 +79,190 @@ local function mirror(target)
 	end
 end
 
+-- Override the list of nameservers (forwarders)
+local function set_nslist(qry, list)
+	for i, ns in ipairs(list) do
+		assert(ffi.C.kr_nsrep_set(qry, i - 1, ns) == 0);
+	end
+	-- If less than maximum NSs, insert guard to terminate the list
+	if #list < 4 then
+		assert(ffi.C.kr_nsrep_set(qry, #list, nil) == 0);
+	end
+end
+
 -- Forward request, and solve as stub query
-local function stub(target)
+function policy.STUB(target)
 	local list = {}
 	if type(target) == 'table' then
 		for _, v in pairs(target) do
-			table.insert(list, addr2sock(v))
+			table.insert(list, addr2sock(v, 53))
 			assert(#list <= 4, 'at most 4 STUB targets are supported')
 		end
 	else
-		table.insert(list, addr2sock(target))
+		table.insert(list, addr2sock(target, 53))
 	end
 	return function(state, req)
 		local qry = req:current()
 		-- Switch mode to stub resolver, do not track origin zone cut since it's not real authority NS
-		qry.flags = bit.band(bit.bor(qry.flags, kres.query.STUB), bit.bnot(kres.query.ALWAYS_CUT))
-		qry:nslist(list)
+		qry.flags.STUB = true
+		qry.flags.ALWAYS_CUT = false
+		set_nslist(qry, list)
 		return state
 	end
 end
 
 -- Forward request and all subrequests to upstream; validate answers
-local function forward(target)
+function policy.FORWARD(target)
 	local list = {}
 	if type(target) == 'table' then
 		for _, v in pairs(target) do
-			table.insert(list, addr2sock(v))
+			table.insert(list, addr2sock(v, 53))
 			assert(#list <= 4, 'at most 4 FORWARD targets are supported')
 		end
 	else
-		table.insert(list, addr2sock(target))
+		table.insert(list, addr2sock(target, 53))
 	end
 	return function(state, req)
 		local qry = req:current()
-		req.options = bit.bor(bit.bor(req.options, kres.query.FORWARD), kres.query.NO_MINIMIZE)
-		qry.flags = bit.band(bit.bor(qry.flags, kres.query.FORWARD), bit.bnot(kres.query.ALWAYS_CUT))
-		qry.flags = bit.bor(qry.flags, kres.query.NO_MINIMIZE)
-		qry.flags = bit.bor(qry.flags, kres.query.AWAIT_CUT)
-		qry:nslist(list)
+		req.options.FORWARD = true
+		req.options.NO_MINIMIZE = true
+		qry.flags.FORWARD = true
+		qry.flags.ALWAYS_CUT = false
+		qry.flags.NO_MINIMIZE = true
+		qry.flags.AWAIT_CUT = true
+		set_nslist(qry, list)
+		return state
+	end
+end
+
+-- object must be non-empty string or non-empty table of non-empty strings
+local function is_nonempty_string_or_table(object)
+	if type(object) == 'string' then
+		return #object ~= 0
+	elseif type(object) ~= 'table' or not next(object) then
+		return false
+	end
+	for _, val in pairs(object) do
+		if type(val) ~= 'string' or #val == 0 then
+			return false
+		end
+	end
+	return true
+end
+
+local function insert_from_string_or_table(source, destination)
+	if type(source) == 'table' then
+		for _, v in pairs(source) do
+			table.insert(destination, v)
+		end
+	else
+		table.insert(destination, source)
+	end
+end
+
+-- Check for allowed authentication types and return type for the current target
+local function tls_forward_target_authtype(idx, target)
+	if (target.pin_sha256 and not (target.ca_file or target.hostname or target.insecure)) then
+		if not is_nonempty_string_or_table(target.pin_sha256) then
+			error('TLS_FORWARD target authentication is invalid at position '
+			      .. idx .. '; pin_sha256 must be string or list of strings')
+		end
+		return 'pin_sha256'
+	elseif (target.insecure and not (target.ca_file or target.hostname or target.pin_sha256)) then
+		return 'insecure'
+	elseif (target.ca_file and target.hostname and not (target.insecure or target.pin_sha256)) then
+		if not (is_nonempty_string_or_table(target.hostname)
+			and is_nonempty_string_or_table(target.ca_file)) then
+			error('TLS_FORWARD target authentication is invalid at position '
+			      .. idx .. '; hostname and ca_file must be string or list of strings')
+		end
+		return 'cert'
+	else
+		error('TLS_FORWARD authentication options at position ' .. idx
+		      .. ' are invalid; specify one of: pin_sha256 / hostname+ca_file / insecure')
+	end
+end
+
+local function tls_forward_target_check_syntax(idx, list_entry)
+	if type(list_entry) ~= 'table' then
+		error('TLS_FORWARD target must be a non-empty table (found '
+		      .. type(list_entry) .. ' at position ' .. idx .. ')')
+	end
+	if type(list_entry[1]) ~= 'string' then
+		error('TLS_FORWARD target must start with an IP address (found '
+		      .. type(list_entry[1]) .. ' at the beginning of target position ' .. idx .. ')')
+	end
+end
+
+-- Forward request and all subrequests to upstream over TLS; validate answers
+function policy.TLS_FORWARD(target)
+	local sockaddr_c_list = {}
+	local sockaddr_config = {}  -- items: { string_addr=<addr string>, auth_type=<auth type> }
+	local ca_files = {}
+	local hostnames = {}
+	local pins = {}
+	if type(target) ~= 'table' or #target < 1 then
+		error('TLS_FORWARD argument must be a non-empty table')
+	end
+	for idx, upstream_list_entry in pairs(target) do
+		tls_forward_target_check_syntax(idx, upstream_list_entry)
+		local auth_type = tls_forward_target_authtype(idx, upstream_list_entry)
+		local string_addr = upstream_list_entry[1]
+		local sockaddr_c = addr2sock(string_addr, 853)
+		local sockaddr_lua = ffi.string(sockaddr_c, ffi.C.kr_sockaddr_len(sockaddr_c))
+		if sockaddr_config[sockaddr_lua] then
+			error('TLS_FORWARD configuration cannot declare two configs for IP address ' .. string_addr)
+		end
+		table.insert(sockaddr_c_list, sockaddr_c)
+		sockaddr_config[sockaddr_lua] = {string_addr=string_addr, auth_type=auth_type}
+		if auth_type == 'cert' then
+			ca_files[sockaddr_lua] = {}
+			hostnames[sockaddr_lua] = {}
+			insert_from_string_or_table(upstream_list_entry.ca_file, ca_files[sockaddr_lua])
+			insert_from_string_or_table(upstream_list_entry.hostname, hostnames[sockaddr_lua])
+		elseif auth_type == 'pin_sha256' then
+			pins[sockaddr_lua] = {}
+			insert_from_string_or_table(upstream_list_entry.pin_sha256, pins[sockaddr_lua])
+		elseif auth_type ~= 'insecure' then
+			-- insecure does nothing, user does not want authentication
+			assert(false, 'unsupported auth_type')
+		end
+	end
+
+	-- Update the global table of authentication data only if all checks above passed
+	for sockaddr_lua, config in pairs(sockaddr_config) do
+		assert(#config.string_addr > 0)
+		if config.auth_type == 'insecure' then
+			net.tls_client(config.string_addr)
+		elseif config.auth_type == 'pin_sha256' then
+			assert(#pins[sockaddr_lua] > 0)
+			net.tls_client(config.string_addr, pins[sockaddr_lua])
+		elseif config.auth_type == 'cert' then
+			assert(#ca_files[sockaddr_lua] > 0)
+			assert(#hostnames[sockaddr_lua] > 0)
+			net.tls_client(config.string_addr, ca_files[sockaddr_lua], hostnames[sockaddr_lua])
+		else
+			assert(false, 'unsupported auth_type')
+		end
+	end
+
+	return function(state, req)
+		local qry = req:current()
+		req.options.FORWARD = true
+		req.options.NO_MINIMIZE = true
+		qry.flags.FORWARD = true
+		qry.flags.ALWAYS_CUT = false
+		qry.flags.NO_MINIMIZE = true
+		qry.flags.AWAIT_CUT = true
+		req.options.TCP = true
+		qry.flags.TCP = true
+		set_nslist(qry, sockaddr_c_list)
 		return state
 	end
 end
 
 -- Rewrite records in packet
-local function reroute(tbl, names)
+function policy.REROUTE(tbl, names)
 	-- Import renumbering rules
 	local ren = require('renumber')
 	local prefixes = {}
@@ -120,10 +274,11 @@ local function reroute(tbl, names)
 end
 
 -- Set and clear some query flags
-local function flags(opts_set, opts_clear)
-	return function(state, req)
+function policy.FLAGS(opts_set, opts_clear)
+	return function(_, req)
 		local qry = req:current()
-		qry.flags = bit.band(bit.bor(qry.flags, opts_set or 0), bit.bnot(opts_clear or 0))
+		ffi.C.kr_qflags_set  (qry.flags, kres.mk_qflags(opts_set   or {}))
+		ffi.C.kr_qflags_clear(qry.flags, kres.mk_qflags(opts_clear or {}))
 		return nil -- chain rule
 	end
 end
@@ -132,14 +287,14 @@ local function mkauth_soa(answer, dname, mname)
 	if mname == nil then
 		mname = dname
 	end
-	return answer:put(dname, 900, answer:qclass(), kres.type.SOA,
-		mname .. '\6nobody\7invalid\0\0\0\0\0\0\0\14\16\0\0\3\132\0\9\58\128\0\0\3\132')
+	return answer:put(dname, 10800, answer:qclass(), kres.type.SOA,
+		mname .. '\6nobody\7invalid\0\0\0\0\1\0\0\14\16\0\0\4\176\0\9\58\128\0\0\42\48')
 end
 
 local dname_localhost = todname('localhost.')
 
 -- Rule for localhost. zone; see RFC6303, sec. 3
-local function localhost(state, req)
+local function localhost(_, req)
 	local qry = req:current()
 	local answer = req.answer
 	ffi.C.kr_pkt_make_auth_header(answer)
@@ -171,7 +326,7 @@ local dname_rev4_localhost_apex = todname('127.in-addr.arpa');
 -- Answer with locally served minimal 127.in-addr.arpa domain, only having
 -- a PTR record in 1.0.0.127.in-addr.arpa, and with 1.0...0.ip6.arpa. zone.
 -- TODO: much of this would better be left to the hints module (or coordinated).
-local function localhost_reversed(state, req)
+local function localhost_reversed(_, req)
 	local qry = req:current()
 	local answer = req.answer
 
@@ -219,26 +374,18 @@ local function localhost_reversed(state, req)
 	return kres.DONE
 end
 
-local policy = {
-	-- Policies
-	PASS = 1, DENY = 2, DROP = 3, TC = 4, QTRACE = 5,
-	FORWARD = forward, STUB = stub, REROUTE = reroute, MIRROR = mirror, FLAGS = flags,
-	-- Special values
-	ANY = 0,
-}
-
 -- All requests
 function policy.all(action)
-	return function(req, query) return action end
+	return function(_, _) return action end
 end
 
 -- Requests which QNAME matches given zone list (i.e. suffix match)
 function policy.suffix(action, zone_list)
-	local AC = require('aho-corasick')
-	local tree = AC.build(zone_list)
-	return function(req, query)
+	local AC = require('ahocorasick')
+	local tree = AC.create(zone_list)
+	return function(_, query)
 		local match = AC.match(tree, query:name(), false)
-		if match[1] ~= nil then
+		if match ~= nil then
 			return action
 		end
 		return nil
@@ -249,7 +396,7 @@ end
 function policy.suffix_common(action, suffix_list, common_suffix)
 	local common_len = string.len(common_suffix)
 	local suffix_count = #suffix_list
-	return function(req, query)
+	return function(_, query)
 		-- Preliminary check
 		local qname = query:name()
 		if not string.find(qname, common_suffix, -common_len, true) then
@@ -268,7 +415,7 @@ end
 
 -- Filter QNAME pattern
 function policy.pattern(action, pattern)
-	return function(req, query)
+	return function(_, query)
 		if string.find(query:name(), pattern) then
 			return action
 		end
@@ -291,34 +438,74 @@ local function rpz_parse(action, path)
 	if not parser:open(path) then error(string.format('failed to parse "%s"', path)) end
 	while parser:parse() do
 		local name = ffi.string(parser.r_owner, parser.r_owner_length)
-		local action = ffi.string(parser.r_data, parser.r_data_length)
-		rules[name] = action_map[action]
+		local name_action = ffi.string(parser.r_data, parser.r_data_length)
+		rules[name] = action_map[name_action]
 		-- Warn when NYI
-		if #name > 1 and not action_map[action] then
+		if #name > 1 and not action_map[name_action] then
 			print(string.format('[ rpz ] %s:%d: unsupported policy action', path, tonumber(parser.line_counter)))
 		end
 	end
 	return rules
 end
 
+-- RPZ policy set
 -- Create RPZ from zone file
-local function rpz_zonefile(action, path)
+function policy.rpz(action, path)
 	local rules = rpz_parse(action, path)
 	collectgarbage()
-	return function(req, query)
+	return function(_, query)
 		local label = query:name()
-		local action = rules[label]
-		while action == nil and string.len(label) > 0 do
+		local rule = rules[label]
+		while rule == nil and string.len(label) > 0 do
 			label = string.sub(label, string.byte(label) + 2)
-			action = rules['\1*'..label]
+			rule = rules['\1*'..label]
 		end
-		return action
+		return rule
 	end
 end
 
--- RPZ policy set
-function policy.rpz(action, path)
-	return rpz_zonefile(action, path)
+function policy.DENY_MSG(msg)
+	if msg and (type(msg) ~= 'string' or #msg >= 255) then
+		error('DENY_MSG: optional msg must be string shorter than 256 characters')
+        end
+
+	return function (_, req)
+		-- Write authority information
+		local answer = req.answer
+		ffi.C.kr_pkt_make_auth_header(answer)
+		answer:rcode(kres.rcode.NXDOMAIN)
+		answer:begin(kres.section.AUTHORITY)
+		mkauth_soa(answer, answer:qname())
+		if msg then
+			answer:begin(kres.section.ADDITIONAL)
+			answer:put('\11explanation\7invalid', 10800, answer:qclass(), kres.type.TXT,
+				   string.char(#msg) .. msg)
+
+		end
+		return kres.DONE
+	end
+end
+policy.DENY = policy.DENY_MSG() -- compatibility with < 2.0
+
+function policy.DROP(_, _)
+	return kres.FAIL
+end
+
+function policy.TC(state, req)
+	local answer = req.answer
+	if answer.max_size ~= 65535 then
+		answer:tc(1) -- ^ Only UDP queries
+		return kres.DONE
+	else
+		return state
+	end
+end
+
+function policy.QTRACE(_, req)
+	local qry = req:current()
+	req.options.TRACE = true
+	qry.flags.TRACE = true
+	return -- this allows to continue iterating over policy list
 end
 
 -- Evaluate packet in given rules to determine policy action
@@ -329,7 +516,7 @@ function policy.evaluate(rules, req, query, state)
 			local action = rule.cb(req, query)
 			if action ~= nil then
 				rule.count = rule.count + 1
-				local next_state = policy.enforce(state, req, action)
+				local next_state = action(state, req)
 				if next_state then    -- Not a chain rule,
 					return next_state -- stop on first match
 				end
@@ -339,35 +526,6 @@ function policy.evaluate(rules, req, query, state)
 	return
 end
 
--- Enforce policy action
-function policy.enforce(state, req, action)
-	if action == policy.DENY then
-		-- Write authority information
-		local answer = req.answer
-		ffi.C.kr_pkt_make_auth_header(answer)
-		answer:rcode(kres.rcode.NXDOMAIN)
-		answer:begin(kres.section.AUTHORITY)
-		mkauth_soa(answer, '\7blocked\0')
-		return kres.DONE
-	elseif action == policy.DROP then
-		return kres.FAIL
-	elseif action == policy.TC then
-		local answer = req.answer
-		if answer.max_size ~= 65535 then
-			answer:tc(1) -- ^ Only UDP queries
-			return kres.DONE
-		end
-	elseif action == policy.QTRACE then
-		local qry = req:current()
-		req.options = bit.bor(req.options, kres.query.TRACE)
-		qry.flags = bit.bor(qry.flags, kres.query.TRACE)
-		return -- this allows to continue iterating over policy list
-	elseif type(action) == 'function' then
-		return action(state, req)
-	end
-	return state
-end
-
 -- Top-down policy list walk until we hit a match
 -- the caller is responsible for reordering policy list
 -- from most specific to least specific.
@@ -376,10 +534,10 @@ end
 policy.layer = {
 	begin = function(state, req)
 		req = kres.request_t(req)
-		return policy.evaluate(policy.rules, req, req:current(), state) or 
+		return policy.evaluate(policy.rules, req, req:current(), state) or
 		       policy.evaluate(policy.special_names, req, req:current(), state) or
 		       state
-	end,	
+	end,
 	finish = function(state, req)
 		req = kres.request_t(req)
 		return policy.evaluate(policy.postrules, req, req:current(), state) or state
@@ -458,7 +616,7 @@ local private_zones = {
 	'100.51.198.in-addr.arpa.',
 	'113.0.203.in-addr.arpa.',
 	'255.255.255.255.in-addr.arpa.',
-	-- RFC7796
+	-- RFC7793
 	'64.100.in-addr.arpa.',
 	'65.100.in-addr.arpa.',
 	'66.100.in-addr.arpa.',
@@ -541,14 +699,22 @@ policy.rules = {}
 policy.postrules = {}
 policy.special_names = {
 	{
-		cb=policy.suffix_common(policy.DENY, private_zones, todname('arpa.')),
+		cb=policy.suffix_common(policy.DENY_MSG(
+			'Blocking is mandated by standards, see references on '
+			.. 'https://www.iana.org/assignments/'
+			.. 'locally-served-dns-zones/locally-served-dns-zones.xhtml'),
+			private_zones, todname('arpa.')),
 		count=0
 	},
 	{
-		cb=policy.suffix(policy.DENY, {
-			todname('test.'),
-			todname('invalid.'),
-			todname('onion.'), -- RFC7686, 2.4
+		cb=policy.suffix(policy.DENY_MSG(
+			'Blocking is mandated by standards, see references on '
+			.. 'https://www.iana.org/assignments/'
+			.. 'special-use-domain-names/special-use-domain-names.xhtml'),
+			{
+				todname('test.'),
+				todname('onion.'),
+				todname('invalid.'),
 			}),
 		count=0
 	},
@@ -563,7 +729,6 @@ policy.special_names = {
 			todname('arpa.')),
 		count=0
 	},
-	
 }
 
 return policy
